#
# Copyright 2022 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from collections.abc import Generator
import logging
import random
from typing import Any, Callable, TypeVar

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic

from aiohomekit.controller.ble.key import DecryptionKey, EncryptionKey
from aiohomekit.exceptions import EncryptionError
from aiohomekit.model.services import ServicesTypes
from aiohomekit.pdu import (
    OpCode,
    PDUStatus,
    decode_pdu,
    decode_pdu_continuation,
    encode_pdu,
)
from aiohomekit.protocol.tlv import TLV

from .bleak import AIOHomeKitBleakClient
from .const import AdditionalParameterTypes
from .structs import BleRequest

logger = logging.getLogger(__name__)

WrapFuncType = TypeVar("WrapFuncType", bound=Callable[..., Any])

DEFAULT_ATTEMPTS = 2
MAX_REASSEMBLY = 50
ATT_HEADER_SIZE = 3
KEY_OVERHEAD_SIZE = 16


class PDUStatusError(Exception):
    """Raised when the PDU status is not success."""

    def __init__(self, status: PDUStatus, message: str) -> None:
        """Initialize a PDUStatusError."""
        super().__init__(message)
        self.status = status


async def ble_request(
    client: AIOHomeKitBleakClient,
    encryption_key: EncryptionKey | None,
    decryption_key: DecryptionKey | None,
    opcode: OpCode,
    handle: BleakGATTCharacteristic,
    iid: int,
    data: bytes | None = None,
) -> tuple[PDUStatus, bytes]:
    """Send a request to the accessory."""
    tid = random.randrange(1, 254)
    await _write_pdu(client, encryption_key, opcode, handle, iid, data, tid)
    return await _read_pdu(client, decryption_key, handle, tid)


async def _write_pdu(
    client: AIOHomeKitBleakClient,
    encryption_key: EncryptionKey | None,
    opcode: OpCode,
    handle: BleakGATTCharacteristic,
    iid: int,
    data: bytes,
    tid: int,
) -> None:
    """Write a PDU to the accessory."""
    fragment_size = client.determine_fragment_size(
        KEY_OVERHEAD_SIZE if encryption_key else 0, handle
    )
    # Wrap data in one or more PDU's split at fragment_size
    # And write each one to the target characteristic handle
    writes = []
    for data in encode_pdu(opcode, tid, iid, data, fragment_size):
        logger.debug("Queuing fragment for write: %s", data)
        if encryption_key:
            data = encryption_key.encrypt(data)
        writes.append(data)

    for write in writes:
        await client.write_gatt_char(handle, write, True)


async def _read_pdu(
    client: AIOHomeKitBleakClient,
    decryption_key: DecryptionKey | None,
    handle: BleakGATTCharacteristic,
    tid: int,
) -> tuple[PDUStatus, bytes]:
    """Read a PDU from a characteristic."""
    data = await client.read_gatt_char(handle)
    if decryption_key:
        data = decryption_key.decrypt(data)
        if data is False:
            raise EncryptionError("Decryption failed")

    logger.debug("Read fragment: %s", data)

    # Validate the PDU header
    status, expected_length, data = decode_pdu(tid, data)

    # If packet is too short then there may be 1 or more continuation
    # packets. Keep reading until we have enough data.
    #
    # Even if the status is failure, we must read the whole
    # data set or the encryption will be out of sync.
    #
    while len(data) < expected_length:
        next = await client.read_gatt_char(handle)
        if decryption_key:
            next = decryption_key.decrypt(next)
            if next is False:
                raise EncryptionError("Decryption failed")
        logger.debug("Read fragment: %s", next)

        data += decode_pdu_continuation(tid, next)

    return status, data


def raise_for_pdu_status(client: BleakClient, pdu_status: PDUStatus) -> None:
    """Raise on non-success PDU status."""
    if pdu_status != PDUStatus.SUCCESS:
        raise PDUStatusError(
            pdu_status.value,
            f"{client.address}: PDU status was not success: {pdu_status.description} ({pdu_status.value})",
        )


def _decode_pdu_tlv_value(
    client: AIOHomeKitBleakClient, pdu_status: PDUStatus, data: bytes
) -> bytes:
    """Decode the TLV value from the PDU."""
    raise_for_pdu_status(client, pdu_status)
    decoded = dict(TLV.decode_bytes(data))
    return decoded[AdditionalParameterTypes.Value.value]


async def char_write(
    client: BleakClient,
    encryption_key: EncryptionKey | None,
    decryption_key: DecryptionKey | None,
    handle: BleakGATTCharacteristic,
    iid: int,
    body: bytes,
) -> bytes:
    """Execute a CHAR_WRITE request."""
    body = BleRequest(expect_response=1, value=body).encode()
    pdu_status, data = await ble_request(
        client, encryption_key, decryption_key, OpCode.CHAR_WRITE, handle, iid, body
    )
    return _decode_pdu_tlv_value(client, pdu_status, data)


async def _pairing_char_write(
    client: AIOHomeKitBleakClient,
    handle: BleakGATTCharacteristic,
    iid: int,
    request: list[tuple[TLV, bytes]],
) -> dict[int, bytes]:
    """Read or write a characteristic value."""
    buffer = bytearray()
    next_write = TLV.encode_list(request)

    for _ in range(MAX_REASSEMBLY):
        data = await char_write(client, None, None, handle, iid, next_write)
        decoded = dict(TLV.decode_bytearray(data))
        if TLV.kTLVType_FragmentLast in decoded:
            logger.debug("%s: Reassembling final fragment", client.address)
            buffer.extend(decoded[TLV.kTLVType_FragmentLast])
            return dict(TLV.decode_bytes(buffer))
        elif TLV.kTLVType_FragmentData in decoded:
            logger.debug("%s: Reassembling fragment", client.address)
            # There is more data, acknowledge the fragment
            # and keep reading
            buffer.extend(decoded[TLV.kTLVType_FragmentData])
            # Acknowledge the fragment
            # We must construct this manually since TLV.encode_bytes
            # current does not know how to encode a 0 length
            next_write = bytes([TLV.kTLVType_FragmentData, 0])
        else:
            return decoded

    raise ValueError(f"Reassembly failed - too many fragments (max: {MAX_REASSEMBLY})")


async def char_read(
    client: AIOHomeKitBleakClient,
    encryption_key: EncryptionKey | None,
    decryption_key: DecryptionKey | None,
    handle: BleakGATTCharacteristic,
    iid: int,
) -> bytes:
    """Execute a CHAR_READ request."""
    pdu_status, data = await ble_request(
        client, encryption_key, decryption_key, OpCode.CHAR_READ, handle, iid
    )
    return _decode_pdu_tlv_value(client, pdu_status, data)


async def drive_pairing_state_machine(
    client: AIOHomeKitBleakClient,
    characteristic: str,
    state_machine: Generator[tuple[list[tuple[TLV, bytes]], list[TLV]], Any, Any],
) -> Any:
    char = client.get_characteristic(ServicesTypes.PAIRING, characteristic)
    iid = await client.get_characteristic_iid(char)

    request, expected = state_machine.send(None)
    while True:
        try:
            decoded = await _pairing_char_write(client, char, iid, request)
            request, expected = state_machine.send(decoded)
        except StopIteration as result:
            return result.value



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Helpers for detecing homekit devices via zeroconf."""
from __future__ import annotations

from abc import abstractmethod
import asyncio
from collections.abc import AsyncIterable
from dataclasses import dataclass
import logging

import async_timeout
from zeroconf import ServiceListener, ServiceStateChange, Zeroconf
from zeroconf.asyncio import AsyncServiceBrowser, AsyncServiceInfo, AsyncZeroconf

from aiohomekit.characteristic_cache import CharacteristicCacheType
from aiohomekit.controller.abstract import (
    AbstractController,
    AbstractDiscovery,
    AbstractPairing,
)
from aiohomekit.exceptions import AccessoryNotFoundError, TransportNotSupportedError
from aiohomekit.model import Categories
from aiohomekit.model.feature_flags import FeatureFlags
from aiohomekit.model.status_flags import StatusFlags

from .debounce import Debouncer

HAP_TYPE_TCP = "_hap._tcp.local."
HAP_TYPE_UDP = "_hap._udp.local."
CLASS_IN = 1
TYPE_PTR = 12

_TIMEOUT_MS = 3000

logger = logging.getLogger(__name__)


@dataclass
class HomeKitService:

    name: str
    id: str
    model: str
    feature_flags: FeatureFlags
    status_flags: StatusFlags
    config_num: int
    state_num: int
    category: Categories
    protocol_version: str

    type: str

    address: str
    addresses: list[str]
    port: int

    @classmethod
    def from_service_info(cls, service: AsyncServiceInfo) -> HomeKitService:
        if not (addresses := service.parsed_addresses()):
            raise ValueError("Invalid HomeKit Zeroconf record: Missing address")

        props: dict[str, str] = {
            k.decode("utf-8").lower(): v.decode("utf-8")
            for (k, v) in service.properties.items()
        }

        if "id" not in props:
            raise ValueError("Invalid HomeKit Zeroconf record: Missing device ID")

        return cls(
            name=service.name.removesuffix(f".{service.type}"),
            id=props["id"].lower(),
            model=props.get("md", ""),
            config_num=int(props.get("c#", 0)),
            state_num=int(props.get("s#", 0)),
            feature_flags=FeatureFlags(int(props.get("ff", 0))),
            status_flags=StatusFlags(int(props.get("sf", 0))),
            category=Categories(int(props.get("ci", 1))),
            protocol_version=props.get("pv", "1.0"),
            type=service.type,
            address=addresses[0],
            addresses=addresses,
            port=service.port,
        )


class ZeroconfServiceListener(ServiceListener):
    """An empty service listener."""

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        pass

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        pass

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        pass


def find_brower_for_hap_type(azc: AsyncZeroconf, hap_type: str) -> AsyncServiceBrowser:
    for browser in azc.zeroconf.listeners:
        if not isinstance(browser, AsyncServiceBrowser):
            continue
        if hap_type not in browser.types:
            continue
        return browser

    raise TransportNotSupportedError(f"There is no zeroconf browser for {hap_type}")


class ZeroconfDiscovery(AbstractDiscovery):

    description: HomeKitService

    def __init__(self, description: HomeKitService):
        self.description = description

    def _update_from_discovery(self, description: HomeKitService):
        self.description = description


class ZeroconfPairing(AbstractPairing):
    def _async_endpoint_changed(self) -> None:
        """The IP and/or port of the accessory has changed."""
        pass

    def _async_description_update(self, description: HomeKitService | None) -> None:
        old_description = self.description

        super()._async_description_update(description)

        if not description:
            return

        endpoint_changed = False
        if not old_description:
            logger.debug("%s: Device rediscovered", self.id)
            endpoint_changed = True
        elif old_description.address != description.address:
            logger.debug(
                "%s: Device IP changed from %s to %s",
                self.id,
                old_description.address,
                description.address,
            )
            endpoint_changed = True
        elif old_description.port != description.port:
            logger.debug(
                "%s: Device port changed from %s to %s",
                self.id,
                old_description.port,
                description.port,
            )
            endpoint_changed = True

        if endpoint_changed:
            self._async_endpoint_changed()


class ZeroconfController(AbstractController):

    """
    Base class for HAP protocols that rely on Zeroconf discovery.
    """

    hap_type: str

    def __init__(
        self,
        char_cache: CharacteristicCacheType,
        zeroconf_instance: AsyncZeroconf,
    ):
        super().__init__(char_cache)
        self._async_zeroconf_instance = zeroconf_instance
        self._waiters: dict[str, list[asyncio.Future]] = {}
        self._debouncers: dict[str, Debouncer] = {}

    async def async_start(self):
        zc = self._async_zeroconf_instance.zeroconf
        if not zc:
            return self

        self._browser = find_brower_for_hap_type(
            self._async_zeroconf_instance, self.hap_type
        )
        self._browser.service_state_changed.register_handler(self._handle_service)

        infos = [
            AsyncServiceInfo(self.hap_type, record.alias)
            for record in zc.cache.get_all_by_details(self.hap_type, TYPE_PTR, CLASS_IN)
        ]

        tasks = []
        for info in infos:
            if info.load_from_cache(self._async_zeroconf_instance.zeroconf):
                self._async_handle_loaded_service_info(info)
            else:
                tasks.append(self._async_handle_service(info))

        if tasks:
            await asyncio.gather(*tasks)

        return self

    def _handle_service(
        self,
        zeroconf: Zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ):
        if service_type != self.hap_type:
            return

        if not (debouncer := self._debouncers.get(name)):
            info = AsyncServiceInfo(service_type, name)

            async def _async_handle_service():
                if info.load_from_cache(self._async_zeroconf_instance.zeroconf):
                    self._async_handle_loaded_service_info(info)
                else:
                    await self._async_handle_service(info)

            debouncer = self._debouncers[name] = Debouncer(
                logger,
                cooldown=0.5,
                immediate=False,
                function=_async_handle_service,
            )

        debouncer.async_trigger()

        if state_change == ServiceStateChange.Removed:
            self._debouncers.pop(name, None)

    async def async_stop(self):
        self._browser.service_state_changed.unregister_handler(self._handle_service)

    async def async_find(
        self, device_id: str, timeout: float = 10.0
    ) -> ZeroconfDiscovery:
        device_id = device_id.lower()

        if discovery := self.discoveries.get(device_id):
            return discovery

        waiters = self._waiters.setdefault(device_id, [])
        waiter = asyncio.Future()
        waiters.append(waiter)

        try:
            async with async_timeout.timeout(timeout):
                if discovery := await waiter:
                    return discovery
        except asyncio.TimeoutError:
            raise AccessoryNotFoundError(
                f"Accessory with device id {device_id} not found"
            )

    async def async_discover(self) -> AsyncIterable[ZeroconfDiscovery]:
        for device in self.discoveries.values():
            yield device

    async def _async_handle_service(self, info: AsyncServiceInfo):
        """Add a device that became visible via zeroconf."""
        # AsyncServiceInfo already tries 3x
        await info.async_request(self._async_zeroconf_instance.zeroconf, _TIMEOUT_MS)
        self._async_handle_loaded_service_info(info)

    def _async_handle_loaded_service_info(self, info: AsyncServiceInfo) -> None:
        """Handle a service info that was discovered via zeroconf."""
        try:
            description = HomeKitService.from_service_info(info)
        except ValueError as e:
            logger.debug("%s: Not a valid homekit device: %s", info.name, e)
            return

        if discovery := self.discoveries.get(description.id):
            discovery._update_from_discovery(description)
        else:
            discovery = self.discoveries[description.id] = self._make_discovery(
                description
            )

        discovery = self.discoveries[description.id] = self._make_discovery(description)

        if pairing := self.pairings.get(description.id):
            logger.debug(
                "%s: Notifying pairing of description update: %s",
                description.id,
                description,
            )
            pairing._async_description_update(description)

        if waiters := self._waiters.pop(description.id, None):
            for waiter in waiters:
                waiter.set_result(discovery)

    @abstractmethod
    def _make_discovery(self, description: HomeKitService) -> AbstractDiscovery:
        pass



#
# Copyright 2022 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from collections.abc import Callable

from bleak.backends.device import BLEDevice
from bleak_retry_connector import (
    BleakAbortedError,
    BleakConnectionError,
    BleakError,
    BleakNotFoundError,
    establish_connection as retry_establish_connection,
)

from aiohomekit.exceptions import AccessoryDisconnectedError, AccessoryNotFoundError

from .bleak import AIOHomeKitBleakClient

MAX_CONNECT_ATTEMPTS = 5


async def establish_connection(
    device: BLEDevice,
    name: str,
    disconnected_callback: Callable[[AIOHomeKitBleakClient], None],
    max_attempts: int | None = None,
    use_services_cache: bool = False,
    ble_device_callback: Callable[[BLEDevice], None] = None,
) -> AIOHomeKitBleakClient:
    """Establish a connection to the accessory."""
    try:
        return await retry_establish_connection(
            AIOHomeKitBleakClient,
            device,
            name,
            disconnected_callback,
            max_attempts=max_attempts or MAX_CONNECT_ATTEMPTS,
            use_services_cache=use_services_cache,
            ble_device_callback=ble_device_callback,
        )
    except (BleakAbortedError, BleakConnectionError) as ex:
        raise AccessoryDisconnectedError(ex) from ex
    except BleakNotFoundError as ex:
        raise AccessoryNotFoundError(ex) from ex
    except BleakError as ex:
        raise AccessoryDisconnectedError(ex) from ex



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "ChaCha20Poly1305Encryptor",
    "ChaCha20Poly1305Decryptor",
    "hkdf_derive",
    "SrpClient",
    "SrpServer",
]

from .chacha20poly1305 import ChaCha20Poly1305Decryptor, ChaCha20Poly1305Encryptor
from .hkdf import hkdf_derive
from .srp import SrpClient, SrpServer



# AUTOGENERATED, DO NOT EDIT

from .const import ThreadNodeCapabilities, ThreadStatus
from .structs import (
    SelectedRTPStreamConfiguration,
    StreamingStatus,
    SupportedAudioStreamConfiguration,
    SupportedRTPConfiguration,
    SupportedVideoStreamConfiguration,
)

characteristics = {
    "000000A6-0000-1000-8000-0026BB765291": {
        "name": "ACCESSORY_FLAGS",
        "description": "Accessory Flags",
        "perms": ["pr", "ev"],
        "format": "uint32",
    },
    "000000B0-0000-1000-8000-0026BB765291": {
        "name": "ACTIVE",
        "description": "Active",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000001-0000-1000-8000-0026BB765291": {
        "name": "ADMINISTRATOR_ONLY_ACCESS",
        "description": "Administrator Only Access",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "00000064-0000-1000-8000-0026BB765291": {
        "name": "AIR_PARTICULATE_DENSITY",
        "description": "Air Particulate Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "00000065-0000-1000-8000-0026BB765291": {
        "name": "AIR_PARTICULATE_SIZE",
        "description": "Air Particulate Size",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000095-0000-1000-8000-0026BB765291": {
        "name": "AIR_QUALITY",
        "description": "Air Quality",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000005-0000-1000-8000-0026BB765291": {
        "name": "AUDIO_FEEDBACK",
        "description": "Audio Feedback",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "00000068-0000-1000-8000-0026BB765291": {
        "name": "BATTERY_LEVEL",
        "description": "Battery Level",
        "perms": ["pr", "ev"],
        "format": "uint8",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "00000008-0000-1000-8000-0026BB765291": {
        "name": "BRIGHTNESS",
        "description": "Brightness",
        "perms": ["pr", "pw", "ev"],
        "format": "int32",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "00000092-0000-1000-8000-0026BB765291": {
        "name": "CARBON_DIOXIDE_DETECTED",
        "description": "Carbon Dioxide Detected",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000093-0000-1000-8000-0026BB765291": {
        "name": "CARBON_DIOXIDE_LEVEL",
        "description": "Carbon Dioxide Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 100000,
        "min_value": 0,
    },
    "00000094-0000-1000-8000-0026BB765291": {
        "name": "CARBON_DIOXIDE_PEAK_LEVEL",
        "description": "Carbon Dioxide Peak Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 100000,
        "min_value": 0,
    },
    "00000069-0000-1000-8000-0026BB765291": {
        "name": "CARBON_MONOXIDE_DETECTED",
        "description": "Carbon Monoxide Detected",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000090-0000-1000-8000-0026BB765291": {
        "name": "CARBON_MONOXIDE_LEVEL",
        "description": "Carbon Monoxide Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 100,
        "min_value": 0,
    },
    "00000091-0000-1000-8000-0026BB765291": {
        "name": "CARBON_MONOXIDE_PEAK_LEVEL",
        "description": "Carbon Monoxide Peak Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 100,
        "min_value": 0,
    },
    "0000008F-0000-1000-8000-0026BB765291": {
        "name": "CHARGING_STATE",
        "description": "Charging State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000CE-0000-1000-8000-0026BB765291": {
        "name": "COLOR_TEMPERATURE",
        "description": "Color Temperature",
        "perms": ["pr", "pw", "ev"],
        "format": "uint32",
        "max_value": 500,
        "min_value": 140,
        "min_step": 1,
    },
    "0000006A-0000-1000-8000-0026BB765291": {
        "name": "CONTACT_SENSOR_STATE",
        "description": "Contact Sensor State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000000D-0000-1000-8000-0026BB765291": {
        "name": "COOLING_THRESHOLD_TEMPERATURE",
        "description": "Cooling Threshold Temperature",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "celsius",
        "max_value": 35,
        "min_value": 10,
        "min_step": 0.1,
    },
    "000000A9-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_AIR_PURIFIER_STATE",
        "description": "Current Air Purifier State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000006B-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_AMBIENT_LIGHT_LEVEL",
        "description": "Current Ambient Light Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "unit": "lux",
        "max_value": 100000,
        "min_value": 0.0001,
    },
    "0000000E-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_DOOR_STATE",
        "description": "Current Door State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000AF-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_FAN_STATE",
        "description": "Current Fan State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000B1-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_HEATER_COOLER_STATE",
        "description": "Current Heater Cooler State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000000F-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_HEATING_COOLING_STATE",
        "description": "Current Heating Cooling State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000006C-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_HORIZONTAL_TILT_ANGLE",
        "description": "Current Horizontal Tilt Angle",
        "perms": ["pr", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "000000B3-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_HUMIDIFIER_DEHUMIDIFIER_STATE",
        "description": "Current Humidifier Dehumidifier State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000006D-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_POSITION",
        "description": "Current Position",
        "perms": ["pr", "ev"],
        "format": "uint8",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "00000010-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_RELATIVE_HUMIDITY",
        "description": "Current Relative Humidity",
        "perms": ["pr", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "000000AA-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_SLAT_STATE",
        "description": "Current Slat State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000011-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_TEMPERATURE",
        "description": "Current Temperature",
        "perms": ["pr", "ev"],
        "format": "float",
        "unit": "celsius",
        "max_value": 100,
        "min_value": 0,
        "min_step": 0.1,
    },
    "000000C1-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_TILT_ANGLE",
        "description": "Current Tilt Angle",
        "perms": ["pr", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "0000006E-0000-1000-8000-0026BB765291": {
        "name": "CURRENT_VERTICAL_TILT_ANGLE",
        "description": "Current Vertical Tilt Angle",
        "perms": ["pr", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "0000011D-0000-1000-8000-0026BB765291": {
        "name": "DIGITAL_ZOOM",
        "description": "Digital Zoom",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
    },
    "000000AC-0000-1000-8000-0026BB765291": {
        "name": "FILTER_CHANGE_INDICATION",
        "description": "Filter Change Indication",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000AB-0000-1000-8000-0026BB765291": {
        "name": "FILTER_LIFE_LEVEL",
        "description": "Filter Life Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 100,
        "min_value": 0,
    },
    "00000052-0000-1000-8000-0026BB765291": {
        "name": "FIRMWARE_REVISION",
        "description": "Firmware Revision",
        "perms": ["pr"],
        "format": "string",
    },
    "00000053-0000-1000-8000-0026BB765291": {
        "name": "HARDWARE_REVISION",
        "description": "Hardware Revision",
        "perms": ["pr"],
        "format": "string",
    },
    "00000012-0000-1000-8000-0026BB765291": {
        "name": "HEATING_THRESHOLD_TEMPERATURE",
        "description": "Heating Threshold Temperature",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "celsius",
        "max_value": 25,
        "min_value": 0,
        "min_step": 0.1,
    },
    "0000006F-0000-1000-8000-0026BB765291": {
        "name": "HOLD_POSITION",
        "description": "Hold Position",
        "perms": ["pw"],
        "format": "bool",
    },
    "00000013-0000-1000-8000-0026BB765291": {
        "name": "HUE",
        "description": "Hue",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "arcdegrees",
        "max_value": 360,
        "min_value": 0,
        "min_step": 1,
    },
    "00000014-0000-1000-8000-0026BB765291": {
        "name": "IDENTIFY",
        "description": "Identify",
        "perms": ["pw"],
        "format": "bool",
    },
    "0000011F-0000-1000-8000-0026BB765291": {
        "name": "IMAGE_MIRRORING",
        "description": "Image Mirroring",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "0000011E-0000-1000-8000-0026BB765291": {
        "name": "IMAGE_ROTATION",
        "description": "Image Rotation",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "arcdegrees",
        "max_value": 270,
        "min_value": 0,
        "min_step": 90,
    },
    "000000D2-0000-1000-8000-0026BB765291": {
        "name": "IN_USE",
        "description": "In Use",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000D6-0000-1000-8000-0026BB765291": {
        "name": "IS_CONFIGURED",
        "description": "Is Configured",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000070-0000-1000-8000-0026BB765291": {
        "name": "LEAK_DETECTED",
        "description": "Leak Detected",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000019-0000-1000-8000-0026BB765291": {
        "name": "LOCK_CONTROL_POINT",
        "description": "Lock Control Point",
        "perms": ["pw"],
        "format": "tlv8",
    },
    "0000001D-0000-1000-8000-0026BB765291": {
        "name": "LOCK_CURRENT_STATE",
        "description": "Lock Current State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000001C-0000-1000-8000-0026BB765291": {
        "name": "LOCK_LAST_KNOWN_ACTION",
        "description": "Lock Last Known Action",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000001A-0000-1000-8000-0026BB765291": {
        "name": "LOCK_MANAGEMENT_AUTO_SECURITY_TIMEOUT",
        "description": "Lock Management Auto Security Timeout",
        "perms": ["pr", "pw", "ev"],
        "format": "uint32",
        "unit": "seconds",
    },
    "000000A7-0000-1000-8000-0026BB765291": {
        "name": "LOCK_PHYSICAL_CONTROLS",
        "description": "Lock Physical Controls",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "0000001E-0000-1000-8000-0026BB765291": {
        "name": "LOCK_TARGET_STATE",
        "description": "Lock Target State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "0000001F-0000-1000-8000-0026BB765291": {
        "name": "LOGS",
        "description": "Logs",
        "perms": ["pr", "ev"],
        "format": "tlv8",
    },
    "00000020-0000-1000-8000-0026BB765291": {
        "name": "MANUFACTURER",
        "description": "Manufacturer",
        "perms": ["pr"],
        "format": "string",
    },
    "00000021-0000-1000-8000-0026BB765291": {
        "name": "MODEL",
        "description": "Model",
        "perms": ["pr"],
        "format": "string",
    },
    "00000022-0000-1000-8000-0026BB765291": {
        "name": "MOTION_DETECTED",
        "description": "Motion Detected",
        "perms": ["pr", "ev"],
        "format": "bool",
    },
    "0000011A-0000-1000-8000-0026BB765291": {
        "name": "MUTE",
        "description": "Mute",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "00000023-0000-1000-8000-0026BB765291": {
        "name": "NAME",
        "description": "Name",
        "perms": ["pr"],
        "format": "string",
    },
    "0000011B-0000-1000-8000-0026BB765291": {
        "name": "NIGHT_VISION",
        "description": "Night Vision",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "000000C4-0000-1000-8000-0026BB765291": {
        "name": "NITROGEN_DIOXIDE_DENSITY",
        "description": "Nitrogen Dioxide Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "00000024-0000-1000-8000-0026BB765291": {
        "name": "OBSTRUCTION_DETECTED",
        "description": "Obstruction Detected",
        "perms": ["pr", "ev"],
        "format": "bool",
    },
    "00000071-0000-1000-8000-0026BB765291": {
        "name": "OCCUPANCY_DETECTED",
        "description": "Occupancy Detected",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000025-0000-1000-8000-0026BB765291": {
        "name": "ON",
        "description": "On",
        "perms": ["pr", "pw", "ev"],
        "format": "bool",
    },
    "0000011C-0000-1000-8000-0026BB765291": {
        "name": "OPTICAL_ZOOM",
        "description": "Optical Zoom",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
    },
    "00000026-0000-1000-8000-0026BB765291": {
        "name": "OUTLET_IN_USE",
        "description": "Outlet In Use",
        "perms": ["pr", "ev"],
        "format": "bool",
    },
    "000000C3-0000-1000-8000-0026BB765291": {
        "name": "OZONE_DENSITY",
        "description": "Ozone Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "0000004C-0000-1000-8000-0026BB765291": {
        "name": "PAIR_SETUP",
        "description": "Pair Setup",
        "perms": ["pr", "pw"],
        "format": "tlv8",
    },
    "0000004E-0000-1000-8000-0026BB765291": {
        "name": "PAIR_VERIFY",
        "description": "Pair Verify",
        "perms": ["pr", "pw"],
        "format": "tlv8",
    },
    "0000004F-0000-1000-8000-0026BB765291": {
        "name": "PAIRING_FEATURES",
        "description": "Pairing Features",
        "perms": ["pr"],
        "format": "uint8",
    },
    "00000050-0000-1000-8000-0026BB765291": {
        "name": "PAIRING_PAIRINGS",
        "description": "Pairing Pairings",
        "perms": ["pr", "pw"],
        "format": "tlv8",
    },
    "000000C7-0000-1000-8000-0026BB765291": {
        "name": "PM10_DENSITY",
        "description": "PM10 Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "000000C6-0000-1000-8000-0026BB765291": {
        "name": "PM2_5_DENSITY",
        "description": "PM2.5 Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "00000072-0000-1000-8000-0026BB765291": {
        "name": "POSITION_STATE",
        "description": "Position State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000D1-0000-1000-8000-0026BB765291": {
        "name": "PROGRAM_MODE",
        "description": "Program Mode",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000073-0000-1000-8000-0026BB765291": {
        "name": "PROGRAMMABLE_SWITCH_EVENT",
        "description": "Programmable Switch Event",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "000000C9-0000-1000-8000-0026BB765291": {
        "name": "RELATIVE_HUMIDITY_DEHUMIDIFIER_THRESHOLD",
        "description": "Relative Humidity Dehumidifier Threshold",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "000000CA-0000-1000-8000-0026BB765291": {
        "name": "RELATIVE_HUMIDITY_HUMIDIFIER_THRESHOLD",
        "description": "Relative Humidity Humidifier Threshold",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "000000D4-0000-1000-8000-0026BB765291": {
        "name": "REMAINING_DURATION",
        "description": "Remaining Duration",
        "perms": ["pr", "ev"],
        "format": "uint32",
        "max_value": 3600,
        "min_value": 0,
        "min_step": 1,
    },
    "000000AD-0000-1000-8000-0026BB765291": {
        "name": "RESET_FILTER_INDICATION",
        "description": "Reset Filter Indication",
        "perms": ["pw"],
        "format": "uint8",
        "max_value": 1,
        "min_value": 1,
        "min_step": 1,
    },
    "00000028-0000-1000-8000-0026BB765291": {
        "name": "ROTATION_DIRECTION",
        "description": "Rotation Direction",
        "perms": ["pr", "pw", "ev"],
        "format": "int32",
    },
    "00000029-0000-1000-8000-0026BB765291": {
        "name": "ROTATION_SPEED",
        "description": "Rotation Speed",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "0000002F-0000-1000-8000-0026BB765291": {
        "name": "SATURATION",
        "description": "Saturation",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "0000008E-0000-1000-8000-0026BB765291": {
        "name": "SECURITY_SYSTEM_ALARM_TYPE",
        "description": "Security System Alarm Type",
        "perms": ["pr", "ev"],
        "format": "uint8",
        "max_value": 1,
        "min_value": 0,
        "min_step": 1,
    },
    "00000066-0000-1000-8000-0026BB765291": {
        "name": "SECURITY_SYSTEM_CURRENT_STATE",
        "description": "Security System Current State",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000067-0000-1000-8000-0026BB765291": {
        "name": "SECURITY_SYSTEM_TARGET_STATE",
        "description": "Security System Target State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000117-0000-1000-8000-0026BB765291": {
        "name": "SELECTED_RTP_STREAM_CONFIGURATION",
        "description": "Selected RTP Stream Configuration",
        "perms": ["pr", "pw"],
        "format": "tlv8",
        "struct": SelectedRTPStreamConfiguration,
    },
    "00000030-0000-1000-8000-0026BB765291": {
        "name": "SERIAL_NUMBER",
        "description": "Serial Number",
        "perms": ["pr"],
        "format": "string",
    },
    "000000CB-0000-1000-8000-0026BB765291": {
        "name": "SERVICE_LABEL_INDEX",
        "description": "Service Label Index",
        "perms": ["pr"],
        "format": "uint8",
        "max_value": 255,
        "min_value": 1,
        "min_step": 1,
    },
    "000000CD-0000-1000-8000-0026BB765291": {
        "name": "SERVICE_LABEL_NAMESPACE",
        "description": "Service Label Namespace",
        "perms": ["pr"],
        "format": "uint8",
    },
    "000000D3-0000-1000-8000-0026BB765291": {
        "name": "SET_DURATION",
        "description": "Set Duration",
        "perms": ["pr", "pw", "ev"],
        "format": "uint32",
        "max_value": 3600,
        "min_value": 0,
        "min_step": 1,
    },
    "00000118-0000-1000-8000-0026BB765291": {
        "name": "SETUP_ENDPOINTS",
        "description": "Setup Endpoints",
        "perms": ["pr", "pw"],
        "format": "tlv8",
    },
    "000000C0-0000-1000-8000-0026BB765291": {
        "name": "SLAT_TYPE",
        "description": "Slat Type",
        "perms": ["pr"],
        "format": "uint8",
    },
    "00000076-0000-1000-8000-0026BB765291": {
        "name": "SMOKE_DETECTED",
        "description": "Smoke Detected",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000075-0000-1000-8000-0026BB765291": {
        "name": "STATUS_ACTIVE",
        "description": "Status Active",
        "perms": ["pr", "ev"],
        "format": "bool",
    },
    "00000077-0000-1000-8000-0026BB765291": {
        "name": "STATUS_FAULT",
        "description": "Status Fault",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000078-0000-1000-8000-0026BB765291": {
        "name": "STATUS_JAMMED",
        "description": "Status Jammed",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000079-0000-1000-8000-0026BB765291": {
        "name": "STATUS_LOW_BATTERY",
        "description": "Status Low Battery",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "0000007A-0000-1000-8000-0026BB765291": {
        "name": "STATUS_TAMPERED",
        "description": "Status Tampered",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000120-0000-1000-8000-0026BB765291": {
        "name": "STREAMING_STATUS",
        "description": "Streaming Status",
        "perms": ["pr", "ev"],
        "format": "tlv8",
        "struct": StreamingStatus,
    },
    "000000C5-0000-1000-8000-0026BB765291": {
        "name": "SULPHUR_DIOXIDE_DENSITY",
        "description": "Sulphur Dioxide Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "00000115-0000-1000-8000-0026BB765291": {
        "name": "SUPPORTED_AUDIO_STREAM_CONFIGURATION",
        "description": "Supported Audio Stream Configuration",
        "perms": ["pr"],
        "format": "tlv8",
        "struct": SupportedAudioStreamConfiguration,
    },
    "00000116-0000-1000-8000-0026BB765291": {
        "name": "SUPPORTED_RTP_CONFIGURATION",
        "description": "Supported RTP Configuration",
        "perms": ["pr"],
        "format": "tlv8",
        "struct": SupportedRTPConfiguration,
        "array": True,
    },
    "00000114-0000-1000-8000-0026BB765291": {
        "name": "SUPPORTED_VIDEO_STREAM_CONFIGURATION",
        "description": "Supported Video Stream Configuration",
        "perms": ["pr"],
        "format": "tlv8",
        "struct": SupportedVideoStreamConfiguration,
    },
    "000000B6-0000-1000-8000-0026BB765291": {
        "name": "SWING_MODE",
        "description": "Swing Mode",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "000000A8-0000-1000-8000-0026BB765291": {
        "name": "TARGET_AIR_PURIFIER_STATE",
        "description": "Target Air Purifier State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "000000AE-0000-1000-8000-0026BB765291": {
        "name": "TARGET_AIR_QUALITY",
        "description": "Target Air Quality",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000032-0000-1000-8000-0026BB765291": {
        "name": "TARGET_DOOR_STATE",
        "description": "Target Door State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "000000BF-0000-1000-8000-0026BB765291": {
        "name": "TARGET_FAN_STATE",
        "description": "Target Fan State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "000000B2-0000-1000-8000-0026BB765291": {
        "name": "TARGET_HEATER_COOLER_STATE",
        "description": "Target Heater Cooler State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000033-0000-1000-8000-0026BB765291": {
        "name": "TARGET_HEATING_COOLING_STATE",
        "description": "Target Heating Cooling State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "0000007B-0000-1000-8000-0026BB765291": {
        "name": "TARGET_HORIZONTAL_TILT_ANGLE",
        "description": "Target Horizontal Tilt Angle",
        "perms": ["pr", "pw", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "000000B4-0000-1000-8000-0026BB765291": {
        "name": "TARGET_HUMIDIFIER_DEHUMIDIFIER_STATE",
        "description": "Target Humidifier Dehumidifier State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "0000007C-0000-1000-8000-0026BB765291": {
        "name": "TARGET_POSITION",
        "description": "Target Position",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "00000034-0000-1000-8000-0026BB765291": {
        "name": "TARGET_RELATIVE_HUMIDITY",
        "description": "Target Relative Humidity",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "000000BE-0000-1000-8000-0026BB765291": {
        "name": "TARGET_SLAT_STATE",
        "description": "Target Slat State",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "00000035-0000-1000-8000-0026BB765291": {
        "name": "TARGET_TEMPERATURE",
        "description": "Target Temperature",
        "perms": ["pr", "pw", "ev"],
        "format": "float",
        "unit": "celsius",
        "max_value": 38,
        "min_value": 10,
        "min_step": 0.1,
    },
    "000000C2-0000-1000-8000-0026BB765291": {
        "name": "TARGET_TILT_ANGLE",
        "description": "Target Tilt Angle",
        "perms": ["pr", "pw", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "0000007D-0000-1000-8000-0026BB765291": {
        "name": "TARGET_VERTICAL_TILT_ANGLE",
        "description": "Target Vertical Tilt Angle",
        "perms": ["pr", "pw", "ev"],
        "format": "int32",
        "unit": "arcdegrees",
        "max_value": 90,
        "min_value": -90,
        "min_step": 1,
    },
    "00000036-0000-1000-8000-0026BB765291": {
        "name": "TEMPERATURE_DISPLAY_UNITS",
        "description": "Temperature Display Units",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
    },
    "000000D5-0000-1000-8000-0026BB765291": {
        "name": "VALVE_TYPE",
        "description": "Valve Type",
        "perms": ["pr", "ev"],
        "format": "uint8",
    },
    "00000037-0000-1000-8000-0026BB765291": {
        "name": "VERSION",
        "description": "Version",
        "perms": ["pr", "ev"],
        "format": "string",
    },
    "000000C8-0000-1000-8000-0026BB765291": {
        "name": "VOC_DENSITY",
        "description": "VOC Density",
        "perms": ["pr", "ev"],
        "format": "float",
        "max_value": 1000,
        "min_value": 0,
        "min_step": 1,
    },
    "00000119-0000-1000-8000-0026BB765291": {
        "name": "VOLUME",
        "description": "Volume",
        "perms": ["pr", "pw", "ev"],
        "format": "uint8",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
        "min_step": 1,
    },
    "000000B5-0000-1000-8000-0026BB765291": {
        "name": "WATER_LEVEL",
        "description": "Water Level",
        "perms": ["pr", "ev"],
        "format": "float",
        "unit": "percentage",
        "max_value": 100,
        "min_value": 0,
    },
    "00000702-0000-1000-8000-0026BB765291": {
        "name": "THREAD_NODE_CAPABILITIES",
        "description": "Thread Node Capabilities",
        "perms": ["pr"],
        "format": "uint16",
        "min_value": 0,
        "max_value": 31,
        "enum": ThreadNodeCapabilities,
    },
    "00000703-0000-1000-8000-0026BB765291": {
        "name": "THREAD_STATUS",
        "description": "Thread Status",
        "perms": ["pr", "ev"],
        "format": "uint16",
        "min_value": 0,
        "max_value": 127,
        "enum": ThreadStatus,
    },
}



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CharacteristicPermissions:
    """
    See table 5-4 page 67
    """

    paired_read = "pr"
    paired_write = "pw"
    events = "ev"
    addition_authorization = "aa"
    timed_write = "tw"
    hidden = "hd"



#
# Copyright 2022 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

import asyncio
from collections.abc import Callable
from datetime import timedelta
import logging
import random
import struct
import time
from typing import TYPE_CHECKING, Any, Iterable, TypeVar, cast
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.exc import BleakError
from bleak_retry_connector import (
    BLEAK_RETRY_EXCEPTIONS as BLEAK_EXCEPTIONS,
    retry_bluetooth_connection_error,
)

from aiohomekit.exceptions import (
    AccessoryDisconnectedError,
    AccessoryNotFoundError,
    AuthenticationError,
    InvalidError,
    UnknownError,
)
from aiohomekit.model import (
    Accessories,
    AccessoriesState,
    Accessory,
    CharacteristicsTypes,
    Transport,
)
from aiohomekit.model.characteristics import Characteristic, CharacteristicPermissions
from aiohomekit.model.services import Service, ServicesTypes
from aiohomekit.pdu import OpCode, PDUStatus, decode_pdu, encode_pdu
from aiohomekit.protocol import get_session_keys
from aiohomekit.protocol.statuscodes import HapStatusCode
from aiohomekit.protocol.tlv import TLV
from aiohomekit.utils import async_create_task
from aiohomekit.uuid import normalize_uuid

from ..abstract import AbstractPairing, AbstractPairingData
from .bleak import AIOHomeKitBleakClient
from .client import (
    PDUStatusError,
    ble_request,
    drive_pairing_state_machine,
    raise_for_pdu_status,
)
from .connection import establish_connection
from .key import BroadcastDecryptionKey, DecryptionKey, EncryptionKey
from .manufacturer_data import HomeKitAdvertisement, HomeKitEncryptedNotification
from .structs import (
    HAP_BLE_CHARACTERISTIC_CONFIGURATION_REQUEST_TLV,
    HAP_BLE_PROTOCOL_CONFIGURATION_REQUEST_TLV,
    HAP_TLV,
    Characteristic as CharacteristicTLV,
    ProtocolParams,
    ProtocolParamsTLV,
    Service as ServiceTLV,
)
from .values import from_bytes, to_bytes

if TYPE_CHECKING:
    from aiohomekit.controller.ble.controller import BleController

logger = logging.getLogger(__name__)

# The discover timeout is how long we will wait for and advertisement to be
# received. If we don't get it in this time we will try again later since
# the scanner is always running anyways.
DISCOVER_TIMEOUT = 10

# Battery powered devices may not broadcast once paired until
# there is an event so we use a long availablity interval.
AVAILABILITY_INTERVAL = 86400 * 7  # 7 days

NEVER_TIME = -AVAILABILITY_INTERVAL


SERVICE_INSTANCE_ID = "E604E95D-A759-4817-87D3-AA005083A0D1"
SERVICE_INSTANCE_ID_UUID = UUID(SERVICE_INSTANCE_ID)

SERVICE_SIGNATURE_UUID = UUID(CharacteristicsTypes.SERVICE_SIGNATURE)

SKIP_SYNC_SERVICES = {
    ServicesTypes.PAIRING,
    ServicesTypes.TRANSFER_TRANSPORT_MANAGEMENT,
}
# These characteristics are not readable unless there has been a write
WRITE_FIRST_REQUIRED_CHARACTERISTICS = {
    "00000131-0000-1000-8000-0026BB765291",  # Setup Data Stream Transport
    "00000117-0000-1000-8000-0026BB765291",  # Selected RTP Stream Configuration
    "00000118-0000-1000-8000-0026BB765291",  # Setup Endpoints
    "00000138-0000-1000-8000-0026BB765291",  # Unknown write first characteristic
    "246912DC-8FA3-82ED-DEA4-9EB91D8FC2EE",  # Unknown Vendor char seen on Belkin Wemo Switch
}
BLE_AID = 1  # The aid for BLE devices is always 1

ENABLE_BROADCAST_PAYLOAD = TLV.encode_list(
    [
        (
            HAP_BLE_CHARACTERISTIC_CONFIGURATION_REQUEST_TLV.kTLVHAPParamCharacteristicConfigurationProperties,
            int.to_bytes(1, 2, "little"),
        ),
        (
            HAP_BLE_CHARACTERISTIC_CONFIGURATION_REQUEST_TLV.kTLVHAPParamCharacteristicConfigurationBroadcastInterval,
            bytes([0x01]),
        ),
    ]
)

GENERATE_BROADCAST_KEY_PAYLOAD = (
    bytes([HAP_BLE_PROTOCOL_CONFIGURATION_REQUEST_TLV.GenerateBroadcastEncryptionKey])
    + b"\x00"
)

GET_ALL_PARAMS_PAYLOAD = (
    bytes([HAP_BLE_PROTOCOL_CONFIGURATION_REQUEST_TLV.GetAllParams]) + b"\x00"
)

WrapFuncType = TypeVar("WrapFuncType", bound=Callable[..., Any])


def operation_lock(func: WrapFuncType) -> WrapFuncType:
    """Define a wrapper to only allow a single operation at a time."""

    async def _async_operation_lock_wrap(
        self: BlePairing, *args: Any, **kwargs: Any
    ) -> None:
        async with self._operation_lock:
            return await func(self, *args, **kwargs)

    return cast(WrapFuncType, _async_operation_lock_wrap)


def restore_connection_and_resume(func: WrapFuncType) -> WrapFuncType:
    """Define a wrapper restore connection, populate data, and then resume when the operation completes."""

    async def _async_restore_and_resume(
        self: BlePairing, *args: Any, **kwargs: Any
    ) -> None:
        """Restore connection, populate data, and then resume when the operation completes."""
        if self._shutdown:
            return
        await self._populate_accessories_and_characteristics()
        try:
            return await func(self, *args, **kwargs)
        finally:
            logger.debug(
                "%s: Finished %s, checking for subscription restore: %s",
                self.name,
                func.__name__,
                self._restore_pending,
            )
            if not self._shutdown and self._restore_pending:
                await self._async_restore_subscriptions()

    return cast(WrapFuncType, _async_restore_and_resume)


class BlePairing(AbstractPairing):
    """
    This represents a paired HomeKit IP accessory.
    """

    description: HomeKitAdvertisement
    controller: BleController

    def __init__(
        self,
        controller: BleController,
        pairing_data: AbstractPairingData,
        device: BLEDevice | None = None,
        client: AIOHomeKitBleakClient | None = None,
        description: HomeKitAdvertisement | None = None,
        ble_advertisement: AdvertisementData | None = None,
    ) -> None:
        self.device = device
        self.ble_advertisement = ble_advertisement
        self.client = client
        self.description = description
        self.pairing_data = pairing_data

        super().__init__(controller, pairing_data)

        self._last_seen = time.monotonic() if description else NEVER_TIME

        # Encryption
        self._derive = None
        self._session_id = None
        self._encryption_key: EncryptionKey | None = None
        self._decryption_key: DecryptionKey | None = None
        self._broadcast_decryption_key: BroadcastDecryptionKey | None = None

        cached_key = self.broadcast_key
        if cached_key:
            self._broadcast_decryption_key = BroadcastDecryptionKey(cached_key)

        # Used to keep track of which characteristics we already started
        # notifications for
        self._notifications: set[int] = set()
        self._broadcast_notifications: set[int] = set()

        # Only allow one attempt to aquire the connection at a time
        self._connection_lock = asyncio.Lock()
        # We don't want to read/write from characteristics in parallel
        # * If 2 coroutines read from the same char at the same time there
        #   would be a race error - a read result could be overwritten by another.
        # * The enc/dec counters are global. Therefore our API's for
        #   a read/write need to be atomic otherwise we end up having
        #   to guess what encryption counter to use for the decrypt
        self._ble_request_lock = asyncio.Lock()
        # Only allow a single operation at at time
        self._operation_lock = asyncio.Lock()
        # Only allow a single attempt to sync config at a time
        self._config_lock = asyncio.Lock()
        # Only subscribe to characteristics one at a time
        self._subscription_lock = asyncio.Lock()
        # Only process disconnected events once
        self._disconnected_events_lock = asyncio.Lock()

        self._tried_to_connect_once = False
        self._restore_pending = False
        self._fetched_gsn_this_session = False

    @property
    def address(self) -> str:
        """Return the address of the device."""
        return (
            self.device.address
            if self.device
            else self.pairing_data["AccessoryAddress"]
        )

    @property
    def name(self) -> str:
        """Return the name of the pairing with the address."""
        if self.description:
            return f"{self.description.name} [{self.address}] (id={self.id})"
        return f"[{self.address}] (id={self.id})"

    @property
    def rssi(self) -> int | None:
        return self.ble_advertisement.rssi if self.ble_advertisement else None

    @property
    def is_connected(self) -> bool:
        return bool(self.client and self.client.is_connected and self._encryption_key)

    @property
    def is_available(self) -> bool:
        """Returns true if the device is currently available."""
        return self._is_available_at_time(time.monotonic())

    @property
    def poll_interval(self) -> timedelta:
        """Returns how often the device should be polled."""
        if any(a.needs_polling for a in self.accessories):
            # Currently only used for devices that have energy data
            return timedelta(minutes=5)
        return timedelta(hours=24)

    def _is_available_at_time(self, monotonic: float) -> bool:
        """Check if we are considered available at the given time."""
        return self.is_connected or monotonic - self._last_seen < AVAILABILITY_INTERVAL

    @property
    def transport(self) -> Transport:
        """The transport used for the connection."""
        return Transport.BLE

    def _async_ble_update(
        self, device: BLEDevice, ble_advertisement: AdvertisementData
    ) -> None:
        """Update the BLE device and ble_advertisement."""
        self.device = device
        self.ble_advertisement = ble_advertisement

    def _async_description_update(
        self, description: HomeKitAdvertisement | None
    ) -> None:
        """Update the description of the accessory."""
        now = time.monotonic()
        was_available = self._is_available_at_time(now)
        self._last_seen = now
        if not was_available:
            self._callback_availability_changed(True)
        if self.description != description:
            logger.debug(
                "%s: Description updated: old=%s new=%s",
                self.name,
                self.description,
                description,
            )

        super()._async_description_update(description)

    async def _async_request(
        self,
        opcode: OpCode,
        char: Characteristic,
        data: bytes | None = None,
        iid: int | None = None,
    ) -> bytes:
        async with self._ble_request_lock:
            return await self._async_request_under_lock(opcode, char, data, iid)

    async def _async_request_under_lock(
        self,
        opcode: OpCode,
        char: Characteristic,
        data: bytes | None = None,
        iid: int | None = None,
    ) -> bytes:
        assert self._ble_request_lock.locked(), "_ble_request_lock Should be locked"
        if not self.client or not self.client.is_connected:
            logger.debug("%s: Client not connected; rssi=%s", self.name, self.rssi)
            raise AccessoryDisconnectedError(f"{self.name} is not connected")

        if char.handle:
            endpoint = self.client.get_characteristic_by_handle(char.handle)
        else:
            endpoint = self.client.get_characteristic(char.service.type, char.type)

        pdu_status, result_data = await ble_request(
            self.client,
            self._encryption_key,
            self._decryption_key,
            opcode,
            endpoint,
            iid if iid is not None else char.iid,
            data,
        )

        if not self.client or not self.client.is_connected:
            logger.debug("%s: Client not connected; rssi=%s", self.name, self.rssi)
            raise AccessoryDisconnectedError(f"{self.name} is not connected")

        raise_for_pdu_status(self.client, pdu_status)
        return result_data

    def _async_disconnected(self, client: AIOHomeKitBleakClient) -> None:
        """Called when bleak disconnects from the accessory closed the connection."""
        logger.debug("%s: Session closed callback: rssi=%s", self.name, self.rssi)
        self._async_reset_connection_state()

    def _async_reset_connection_state(self) -> None:
        """Reset the connection state after a disconnect."""
        self._encryption_key = None
        self._decryption_key = None
        self._notifications = set()
        self._broadcast_notifications = set()
        self._restore_pending = False
        self._fetched_gsn_this_session = False

    async def _ensure_connected(self, attempts: int | None = None) -> bool | None:
        """Ensure that we are connected to the accessory.

        Returns True if we had to make the connection,
        returns False if we were already connected or shutdown.
        """
        assert self._config_lock.locked(), "_config_lock Should be locked"
        if self._shutdown or (self.client and self.client.is_connected):
            return False
        async with self._connection_lock:
            # Check again while holding the lock
            if self._shutdown or (self.client and self.client.is_connected):
                return False
            if not self.device and (
                discovery := await self.controller.async_get_discovery(
                    self.address, DISCOVER_TIMEOUT
                )
            ):
                self.device = discovery.device
                self.ble_advertisement = discovery.ble_advertisement
                self.description = discovery.description
            elif not self.device:
                raise AccessoryNotFoundError(
                    f"{self.name}: Could not find {self.address}"
                )
            self.client = await establish_connection(
                self.device,
                self.name,
                self._async_disconnected,
                use_services_cache=True,
                ble_device_callback=lambda: self.device,
                max_attempts=attempts,
            )
            return True

    async def _async_start_notify(self, iid: int) -> None:
        char = self.accessories.aid(BLE_AID).characteristics.iid(iid)

        # Find the GATT Characteristic object for this iid
        if char.handle:
            endpoint = self.client.get_characteristic_by_handle(char.handle)
        else:
            endpoint = self.client.get_characteristic(char.service.type, char.type)

        # We only want to allow one in flight read
        # and one pending read at a time since there
        # may be a notify storm and the read it always
        # going to give us the latest value anyways
        max_callback_enforcer = asyncio.Semaphore(2)

        async def _async_callback() -> None:
            if max_callback_enforcer.locked():
                # Already one being read now, and one pending
                return
            async with max_callback_enforcer:
                if not self.client or not self.client.is_connected:
                    # Client disconnected
                    return
                logger.debug("%s: Retrieving event for iid: %s", self.name, iid)
                await self._get_characteristics_without_retry(
                    [(BLE_AID, iid)], notify_listeners=True
                )
                # After a char has changed we need to check if the
                # GSN has changed as well so we don't reconnect
                # to the accessory if we don't need to
                if self._fetched_gsn_this_session:
                    # The spec says we should only do this once per session
                    # after a characteristic has changed
                    return
                protocol_param = await self._get_all_protocol_params()
                if protocol_param:
                    self.description.state_num = protocol_param.state_number
                    self._fetched_gsn_this_session = True

        def _callback(id: int, data: bytes) -> None:
            logger.debug("%s: Received event for iid=%s: %s", self.name, iid, data)
            if data != b"":
                # We should only poll on empty messages, otherwise we may poll
                # the device every second on DBUS systems.
                return
            if max_callback_enforcer.locked():
                # Already one being read now, and one pending
                return
            async_create_task(_async_callback())

        logger.debug("%s: Subscribing to gatt notify for iid: %s", self.name, iid)
        await self.client.start_notify(endpoint, _callback)
        self._notifications.add(iid)

    async def _async_pair_verify(self) -> None:
        async with self._ble_request_lock:
            session_id, derive = await drive_pairing_state_machine(
                self.client,
                CharacteristicsTypes.PAIR_VERIFY,
                get_session_keys(self.pairing_data, self._session_id, self._derive),
            )
            self._encryption_key = EncryptionKey(
                derive(b"Control-Salt", b"Control-Write-Encryption-Key")
            )
            self._decryption_key = DecryptionKey(
                derive(b"Control-Salt", b"Control-Read-Encryption-Key")
            )
            # Used for session resume
            self._session_id = session_id
            self._derive = derive

    async def _process_disconnected_events(self) -> None:
        """Handle disconnected events seen from the advertisement."""
        if not self._tried_to_connect_once:
            # We never tried connected to the accessory, so we don't need to
            # process the disconnected events
            logger.debug(
                "%s: Skipping disconnected events because we have not yet connected.",
                self.name,
            )
            return

        if self._disconnected_events_lock.locked():
            # Already processing disconnected events
            return

        async with self._disconnected_events_lock:
            logger.debug(
                "%s: Polling subscriptions for changes during disconnection; rssi=%s",
                self.name,
                self.rssi,
            )
            try:
                protocol_param = await self._process_disconnected_events_with_retry()
            except (
                AccessoryDisconnectedError,
                *BLEAK_EXCEPTIONS,
                AccessoryNotFoundError,
            ) as exc:
                logger.exception(
                    "%s: Failed to fetch disconnected events: %s; rssi=%s",
                    self.name,
                    exc,
                    self.rssi,
                )
                return

            if protocol_param:
                self.description.state_num = protocol_param.state_number

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def _process_disconnected_events_with_retry(
        self,
    ) -> ProtocolParams | None:
        accessory_chars = self.accessories.aid(BLE_AID).characteristics
        protocol_param = await self._get_all_protocol_params()
        chars_to_update = []
        for _, iid in self.subscriptions:
            char = accessory_chars.iid(iid)
            if char.broadcast_events or char.disconnected_events:
                chars_to_update.append(char)
        if chars_to_update:
            await self._get_characteristics_while_connected(
                chars_to_update,
                notify_listeners=True,
            )
        return protocol_param

    def _async_notification(self, data: HomeKitEncryptedNotification) -> None:
        """Receive a notification from the accessory."""
        if not self._broadcast_decryption_key:
            logger.debug(
                "%s: Received notification before session is setup, "
                "falling back processing as disconnected event: %s",
                self.name,
                data,
            )
            async_create_task(self._process_disconnected_events())
            return

        start_state_num = self.description.state_num
        # Usually we increment by one, but sometimes we get multiple with the same
        # state number so the first pass is optimistic to reduce the number of
        # of decrypts we do.
        for state_num in (
            start_state_num + 1,
            start_state_num,
            *range(start_state_num + 2, start_state_num + 30),
        ):
            logger.debug(
                "%s: Trying state_num %s for encrypted notification: %s",
                self.name,
                state_num,
                data,
            )
            decrypted = self._broadcast_decryption_key.decrypt(
                data.encrypted_payload,
                state_num,
                data.advertising_identifier,
            )
            if decrypted is None:
                continue
            gsn = int.from_bytes(decrypted[0:2], "little")
            if gsn != state_num:
                logger.debug(
                    "%s: GSN mismatch, expected: %s, got: %s",
                    self.name,
                    state_num,
                    gsn,
                )
                return
            iid = int.from_bytes(decrypted[2:4], "little")
            value = decrypted[4:12]
            logger.debug(
                "%s: Received notification: encrypted =  %s - decrypted = %s - gsn=%s - iid=%s - value=%s",
                self.name,
                data.encrypted_payload,
                decrypted,
                gsn,
                iid,
                value,
            )
            # We had a successful decrypt, so we can update the state_num
            self.description.state_num = gsn
            char = self.accessories.aid(BLE_AID).characteristics.iid(iid)

            results = {(BLE_AID, iid): {"value": from_bytes(char, value)}}
            logger.debug("%s: Received notification: results = %s", self.name, results)

            for listener in self.listeners:
                listener(results)
            return

        logger.warning(
            "%s: Received notification but could not decrypt: %s", self.name, data
        )

    def _async_get_service_signature_char(self) -> Characteristic | None:
        """Get the service signature characteristic."""
        info = self.accessories.aid(BLE_AID).services.first(
            service_type=ServicesTypes.PROTOCOL_INFORMATION
        )
        if not info:
            logger.debug("%s: No signature service found", self.name)
            return None
        if not info.has(CharacteristicsTypes.SERVICE_SIGNATURE):
            logger.debug(
                "%s: No signature characteristic found, "
                "accessory may not implement encrypted notifications",
                self.name,
            )
            return None
        return info[CharacteristicsTypes.SERVICE_SIGNATURE]

    async def _async_set_broadcast_encryption_key(self) -> None:
        """Get the broadcast key for the accessory."""
        logger.debug("%s: Setting broadcast encryption key", self.name)
        if self._ble_request_lock.locked():
            logger.debug(
                "%s: Waiting ble request lock to set broadcast encryption key",
                self.name,
            )
        async with self._ble_request_lock:
            hap_char = self._async_get_service_signature_char()
            if not hap_char:
                return
            service_iid = hap_char.service.iid
            logger.debug(
                "%s: Setting broadcast key for service_iid: %s",
                self.name,
                service_iid,
            )
            try:
                await self._async_request_under_lock(
                    OpCode.PROTOCOL_CONFIG,
                    hap_char,
                    GENERATE_BROADCAST_KEY_PAYLOAD,
                    iid=service_iid,
                )
            except PDUStatusError:
                logger.exception(
                    "%s: Failed to set broadcast key, try un-paring and re-pairing the accessory.",
                    self.name,
                )
        long_term_pub_key_hex: str = self.pairing_data["iOSDeviceLTPK"]
        long_term_pub_key_bytes = bytes.fromhex(long_term_pub_key_hex)
        broadcast_key_bytes = self._derive(
            long_term_pub_key_bytes, b"Broadcast-Encryption-Key"
        )
        self._broadcast_decryption_key = BroadcastDecryptionKey(broadcast_key_bytes)
        if self._accessories_state and self.broadcast_key != broadcast_key_bytes:
            self._accessories_state.broadcast_key = broadcast_key_bytes
            self._callback_and_save_config_changed(self.config_num)

    async def _read_signature(
        self,
        char: BleakGATTCharacteristic,
        op_code: OpCode,
        iid: int,
        tlv_struct: ServiceTLV | CharacteristicTLV,
    ) -> dict[str, Any]:
        """Read the signature for the given characteristic."""
        tid = random.randint(1, 254)
        for data in encode_pdu(op_code, tid, iid):
            await self.client.write_gatt_char(
                char,
                data,
                "write-without-response" not in char.properties,
            )

        payload = await self.client.read_gatt_char(char)

        status, _, signature = decode_pdu(tid, payload)
        if status != PDUStatus.SUCCESS:
            return {}

        return tlv_struct.decode(signature).to_dict()

    async def _async_fetch_gatt_database(self) -> Accessories:
        logger.debug("%s: Fetching GATT database; rssi=%s", self.name, self.rssi)
        accessory = Accessory()
        accessory.aid = 1
        # Never use the cache when fetching the GATT database
        services_to_link: list[tuple[Service, list[int]]] = []
        services = await self.client.get_services()
        for service in services:
            ble_service_char = service.get_characteristic(SERVICE_INSTANCE_ID_UUID)
            if not ble_service_char:
                logger.debug(
                    "%s: Skipping service without service instance id: %s",
                    self.name,
                    service,
                )
                continue

            service_iid_bytes = await self.client.read_gatt_char(
                ble_service_char.handle
            )
            service_iid = int.from_bytes(service_iid_bytes, "little")
            logger.debug(
                "%s: Service %s iid: %s (decoded as %s)",
                self.name,
                service.uuid,
                service_iid_bytes,
                service_iid,
            )
            s = accessory.add_service(normalize_uuid(service.uuid))
            s.iid = service_iid

            service_signature_char = service.get_characteristic(SERVICE_SIGNATURE_UUID)
            if service_signature_char:
                decoded_service = await self._read_signature(
                    service_signature_char,
                    OpCode.SERV_SIG_READ,
                    service_iid,
                    ServiceTLV,
                )
                if "linked" in decoded_service:
                    services_to_link.append((s, decoded_service["linked"]))
                logger.debug(
                    "%s: service: %s decoded: %s", self.name, service, decoded_service
                )

            for char in service.characteristics:
                normalized_uuid = normalize_uuid(char.uuid)
                if normalized_uuid == SERVICE_INSTANCE_ID:
                    continue

                iid = await self.client.get_characteristic_iid(char)
                if iid is None:
                    logger.debug("%s: No iid for %s", self.name, char.uuid)
                    continue

                decoded = await self._read_signature(
                    char, OpCode.CHAR_SIG_READ, iid, CharacteristicTLV
                )
                if normalized_uuid == CharacteristicsTypes.IDENTIFY:
                    # Workaround for older eve v1 devices which has a broken identify characteristic
                    # that presents identify as data.
                    decoded["format"] = "bool"

                hap_char = s.add_char(normalized_uuid)
                logger.debug("%s: char: %s decoded: %s", self.name, char, decoded)

                hap_char.iid = iid
                hap_char.handle = char.handle
                hap_char.perms = decoded["perms"]
                # Some vendor characteristics have no format
                # See https://github.com/home-assistant/core/issues/76104
                if "format" in decoded:
                    hap_char.format = decoded["format"]
                if "minStep" in decoded:
                    hap_char.minStep = decoded["minStep"]
                if "minValue" in decoded:
                    hap_char.minValue = decoded["minValue"]
                if "maxValue" in decoded:
                    hap_char.maxValue = decoded["maxValue"]
                if "disconnected_events" in decoded:
                    hap_char.disconnected_events = decoded["disconnected_events"]
                if "broadcast_events" in decoded:
                    hap_char.broadcast_events = decoded["broadcast_events"]

        # Link services after we are done since we need to have all services
        # in the accessory to link them.
        for service, linked in services_to_link:
            for iid in linked:
                service.add_linked_service(accessory.services.iid(iid))

        accessories = Accessories()
        accessories.add_accessory(accessory)
        logger.debug("%s: Completed fetching GATT database", self.name)

        return accessories

    @operation_lock
    async def close_after_operation(self) -> None:
        """Close the client after an operation."""
        await self.close()

    async def close(self) -> None:
        async with self._connection_lock:
            await self._close_while_locked()

    async def _close_while_locked(self) -> None:
        if not self.client or not self.client.is_connected:
            return
        try:
            await self.client.disconnect()
        except BleakError:
            logger.debug(
                "%s: Failed to close connection, client may have already closed it; rssi=%s",
                self.name,
                self.rssi,
            )
        self.client = None
        self._async_reset_connection_state()
        logger.debug(
            "%s: Connection closed from close call; rssi=%s", self.name, self.rssi
        )

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def list_accessories_and_characteristics(self) -> list[dict[str, Any]]:
        return self.accessories.serialize()

    async def _populate_char_values(self, config_changed: bool) -> None:
        """Populate the values of all characteristics."""
        chars: list[Characteristic] = []
        for service in self.accessories.aid(BLE_AID).services:
            if service.type in SKIP_SYNC_SERVICES:
                continue
            if (
                not config_changed
                and service.type == ServicesTypes.ACCESSORY_INFORMATION
            ):
                continue
            for char in service.characteristics:
                if char.type in WRITE_FIRST_REQUIRED_CHARACTERISTICS:
                    continue
                if CharacteristicPermissions.paired_read not in char.perms:
                    continue
                chars.append(char)

        if not chars:
            return

        protocol_params = await self._get_all_protocol_params()

        results = await self._get_characteristics_while_connected(chars)
        logger.debug("%s: Read %s", self.name, results)
        for char in chars:
            result = results.get((BLE_AID, char.iid))
            if not result or "value" not in result:
                logger.debug("%s: No value for %s", self.name, char)
                continue
            char.value = result["value"]

        if protocol_params:
            self.description.state_num = protocol_params.state_number

    async def _get_all_protocol_params(self) -> ProtocolParams | None:
        """Get the current protocol params number."""
        hap_char = self._async_get_service_signature_char()
        if not hap_char:
            return
        service_iid = hap_char.service.iid
        try:
            resp = await self._async_request(
                OpCode.PROTOCOL_CONFIG,
                hap_char,
                GET_ALL_PARAMS_PAYLOAD,
                iid=service_iid,
            )
        except PDUStatusError:
            logger.exception(
                "%s: Failed to get global state number.",
                self.name,
            )
            return None
        response = dict(TLV.decode_bytes(resp))
        protocol_params = ProtocolParams(
            state_number=int.from_bytes(
                response[ProtocolParamsTLV.GlobalStateNumber], "little"
            ),
            config_number=int.from_bytes(
                response[ProtocolParamsTLV.ConfigurationNumber], "little"
            ),
            advertising_id=response[ProtocolParamsTLV.AdvertisingId],
            broadcast_key=response.get(ProtocolParamsTLV.BroadcastKey),
        )
        logger.debug(
            "%s: Fetched protocol params: gsn=%s, c#=%s",
            self.name,
            protocol_params.state_number,
            protocol_params.config_number,
        )
        return protocol_params

    async def async_populate_accessories_state(
        self, force_update: bool = False, attempts: int | None = None
    ) -> None:
        """Populate the state of all accessories.

        This method should try not to fetch all the accessories unless
        we know the config num is out of date.

        Callers should not get BleakError as they expect to trap
        AccessoryDisconnectedError.
        """
        try:
            await self._async_populate_accessories_state(force_update, attempts)
        except BleakError as ex:
            raise AccessoryDisconnectedError(
                f"{self.name} connection failed: {ex}; rssi={self.rssi}"
            ) from ex

    @operation_lock
    @retry_bluetooth_connection_error()
    async def _async_populate_accessories_state(
        self, force_update: bool = False, attempts: int | None = None
    ) -> None:
        """Populate the state of all accessories under the lock."""
        await self._populate_accessories_and_characteristics(force_update, attempts)
        if self._restore_pending:
            await self._async_restore_subscriptions()

    def _all_handles_are_missing(self) -> bool:
        """Check if any characteristic has a handle.

        Older code did not save the handle, so if we have no handles
        we need to re-fetch the gatt database.
        """
        for service in self.accessories.aid(BLE_AID).services:
            for char in service.characteristics:
                if char.handle is not None:
                    return False
        return True

    async def _populate_accessories_and_characteristics(
        self, force_update: bool = False, attempts: int | None = None
    ) -> None:
        was_locked = self._config_lock.locked()
        async with self._config_lock:
            if self._shutdown:
                return

            try:
                made_connection = await self._ensure_connected(attempts)
            finally:
                # Only set _tried_to_connect_once after the connection
                # attempt is complete so we don't try to process disconnected
                # events while we are still trying to connect.
                self._tried_to_connect_once = True

            logger.debug(
                "%s: Populating accessories and characteristics: made_connection=%s restore_pending=%s",
                self.name,
                made_connection,
                self._restore_pending,
            )
            self._restore_pending |= made_connection

            if was_locked and not force_update:
                # No need to do it twice if we already have the data
                # and we are not forcing an update
                return

            update_values = force_update
            config_changed = False
            if self.description:
                config_changed = self.config_num != self.description.config_num

            if (
                not self.accessories
                or config_changed
                or self._all_handles_are_missing()
            ):
                logger.debug(
                    "%s: Fetching gatt database because, cached_config_num: %s, adv config_num: %s",
                    self.name,
                    self.config_num,
                    self.description.config_num,
                )
                accessories = await self._async_fetch_gatt_database()
                new_config_num = self.description.config_num if self.description else 0
                self._accessories_state = AccessoriesState(
                    accessories, new_config_num, self.broadcast_key
                )
                update_values = True

            if not self._encryption_key:
                await self._async_pair_verify()

            if update_values:
                await self._populate_char_values(config_changed)
                self._update_accessories_state_cache()

            if config_changed:
                self._callback_and_save_config_changed(self.config_num)

    async def _async_subscribe_broadcast_events(
        self, subscriptions: list[tuple[int, int]]
    ) -> None:
        """Subscribe to broadcast events."""
        accessory_chars = self.accessories.aid(BLE_AID).characteristics
        to_subscribe: list[Characteristic] = []
        for _, iid in subscriptions:
            hap_char = accessory_chars.iid(iid)
            if (
                not hap_char
                or not hap_char.broadcast_events
                or iid in self._broadcast_notifications
            ):
                continue
            to_subscribe.append(hap_char)

        if not to_subscribe:
            return

        async with self._ble_request_lock:
            for hap_char in to_subscribe:
                iid = hap_char.iid
                if iid in self._broadcast_notifications:
                    continue  # check again with the lock
                logger.debug(
                    "%s: Subscribing to broadcast notify for iid: %s", self.name, iid
                )
                try:
                    await self._async_request_under_lock(
                        OpCode.CHAR_CONFIG, hap_char, ENABLE_BROADCAST_PAYLOAD
                    )
                except PDUStatusError:
                    logger.debug(
                        "%s: Failed to subscribe to broadcast events for %s",
                        self.name,
                        hap_char,
                    )
                    continue
                self._broadcast_notifications.add(iid)

    async def _async_restore_subscriptions(self) -> None:
        """Restore subscriptions and setup notifications after after connecting."""
        if not self._restore_pending or not self.client or not self.client.is_connected:
            return

        if not self.subscriptions:
            logger.debug("%s: No subscriptions to restore", self.name)
            self._restore_pending = False
            return

        await self._async_set_broadcast_encryption_key()
        subscriptions = list(self.subscriptions)
        logger.debug(
            "%s: Connected, resuming subscriptions: %s; rssi=%s",
            self.name,
            subscriptions,
            self.rssi,
        )
        await self._async_subscribe_broadcast_events(subscriptions)
        await self._async_start_notify_subscriptions(subscriptions)
        self._restore_pending = False
        # After we have restored subscriptions, we need to read
        # the state number again to make sure we are in sync
        protocol_param = await self._get_all_protocol_params()
        if protocol_param:
            self.description.state_num = protocol_param.state_number

    async def _async_start_notify_subscriptions(
        self, subscriptions: list[tuple[int, int]]
    ) -> None:
        """Start notifications for the given subscriptions."""
        for _, iid in subscriptions:
            if iid in self._notifications:
                continue
            # The iid will not be in in self._notifications until
            # the _async_start_notify call returns.
            async with self._subscription_lock:
                if iid in self._notifications or not self.client.is_connected:
                    continue
                try:
                    await self._async_start_notify(iid)
                except BLEAK_EXCEPTIONS as ex:
                    # Likely disconnected before we could start notifications
                    # we will get disconnected events instead.
                    logger.debug(
                        "%s: Could not start notify for %s: %s; rssi=%s",
                        self.name,
                        iid,
                        ex,
                        self.rssi,
                    )

    @operation_lock
    @retry_bluetooth_connection_error()
    async def _process_config_changed(self, config_num: int) -> None:
        """Process a config change.

        This method is called when the config num changes.
        """
        await self._populate_accessories_and_characteristics()
        if self._restore_pending:
            await self._async_restore_subscriptions()

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def list_pairings(self):
        request_tlv = TLV.encode_list(
            [(TLV.kTLVType_State, TLV.M1), (TLV.kTLVType_Method, TLV.ListPairings)]
        )
        request_tlv = TLV.encode_list(
            [
                (TLV.kTLVHAPParamParamReturnResponse, bytearray(b"\x01")),
                (TLV.kTLVHAPParamValue, request_tlv),
            ]
        )

        info = self.accessories.aid(BLE_AID).services.first(
            service_type=ServicesTypes.PAIRING
        )
        char = info[CharacteristicsTypes.PAIRING_PAIRINGS]

        resp = await self._async_request(OpCode.CHAR_WRITE, char, request_tlv)

        response = dict(TLV.decode_bytes(resp))

        resp = TLV.decode_bytes(response[1])

        tmp = []
        r = {}
        for d in resp[1:]:
            if d[0] == TLV.kTLVType_Identifier:
                r = {}
                tmp.append(r)
                r["pairingId"] = d[1].decode()
            if d[0] == TLV.kTLVType_PublicKey:
                r["publicKey"] = d[1].hex()
            if d[0] == TLV.kTLVType_Permissions:
                controller_type = "regular"
                if d[1] == b"\x01":
                    controller_type = "admin"
                r["permissions"] = int.from_bytes(d[1], byteorder="little")
                r["controllerType"] = controller_type
        return tmp

    @retry_bluetooth_connection_error()
    async def get_characteristics(
        self,
        characteristics: list[tuple[int, int]],
    ) -> dict[tuple[int, int], dict[str, Any]]:
        return await self._get_characteristics_without_retry(characteristics)

    @operation_lock
    @restore_connection_and_resume
    async def _get_characteristics_without_retry(
        self,
        characteristics: list[tuple[int, int]],
        notify_listeners: bool = False,
    ) -> dict[tuple[int, int], dict[str, Any]]:
        accessory_chars = self.accessories.aid(BLE_AID).characteristics
        return await self._get_characteristics_while_connected(
            [accessory_chars.iid(iid) for _, iid in characteristics], notify_listeners
        )

    async def _get_characteristics_while_connected(
        self,
        characteristics: list[Characteristic],
        notify_listeners: bool = False,
    ) -> dict[tuple[int, int], dict[str, Any]]:
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "%s: Reading characteristics: %s; rssi=%s",
                self.name,
                [char.iid for char in characteristics],
                self.rssi,
            )

        results = {}

        async with self._ble_request_lock:
            for char in characteristics:
                if char.type in WRITE_FIRST_REQUIRED_CHARACTERISTICS:
                    logger.debug(
                        "%s: Ignoring write first only characteristic %s",
                        self.name,
                        char.iid,
                    )
                    continue

                logger.debug("%s: Reading characteristic %s", self.name, char.type)

                try:
                    data = await self._async_request_under_lock(OpCode.CHAR_READ, char)
                except PDUStatusError as ex:
                    # For the apple defines ones we know about we can avoid triggering
                    # this state, but we do not know all the vendor custom chars so
                    # we need to skip in this case.
                    if ex.status == PDUStatus.INVALID_REQUEST:
                        logger.debug(
                            "%s: Reading characteristic %s resulted in an invalid request (skipped)",
                            self.name,
                            char.type,
                        )
                        continue
                decoded = dict(TLV.decode_bytes(data))[1]

                logger.debug(
                    "%s: Read characteristic %s got data, expected format is %s: data=%s decoded=%s",
                    self.name,
                    char.type,
                    char.format,
                    data,
                    decoded,
                )

                try:
                    value = from_bytes(char, decoded)
                except struct.error as ex:
                    logger.debug(
                        "%s: Failed to decode characteristic for %s from %s: %s",
                        self.name,
                        char,
                        decoded,
                        ex,
                    )
                    continue

                result_key = (BLE_AID, char.iid)
                result_value = {"value": value}
                results[result_key] = result_value

                if notify_listeners:
                    # Since it can take a while to read all the characteristics
                    # we want to notify the listeners as soon as we have the
                    # value for each characteristic.
                    single_results = {result_key: result_value}
                    for listener in self.listeners:
                        listener(single_results)

        return results

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def put_characteristics(
        self, characteristics: list[tuple[int, int, Any]]
    ) -> dict[tuple[int, int], Any]:
        results: dict[tuple[int, int], Any] = {}
        logger.debug(
            "%s: Writing characteristics: %s; rssi=%s",
            self.name,
            characteristics,
            self.rssi,
        )
        accessory_chars = self.accessories.aid(BLE_AID).characteristics
        async with self._ble_request_lock:

            for aid, iid, value in characteristics:
                char = accessory_chars.iid(iid)
                logger.debug(
                    "%s: Writing characteristics: iid=%s value=%s",
                    self.name,
                    iid,
                    value,
                )

                if CharacteristicPermissions.timed_write in char.perms:
                    payload_inner = TLV.encode_list(
                        [
                            (HAP_TLV.kTLVHAPParamValue, to_bytes(char, value)),
                            (HAP_TLV.kTLVHAPParamTTL, b"\x1e"),  # 3.0s
                        ]
                    )
                    payload = (len(payload_inner)).to_bytes(
                        length=2, byteorder="little"
                    ) + payload_inner
                    await self._async_request_under_lock(
                        OpCode.CHAR_TIMED_WRITE, char, payload
                    )
                    await self._async_request_under_lock(OpCode.CHAR_EXEC_WRITE, char)

                elif CharacteristicPermissions.paired_write in char.perms:
                    payload = TLV.encode_list(
                        [(HAP_TLV.kTLVHAPParamValue, to_bytes(char, value))]
                    )
                    await self._async_request_under_lock(
                        OpCode.CHAR_WRITE, char, payload
                    )

                else:
                    results[(aid, iid)] = {
                        "status": HapStatusCode.CANT_WRITE_READ_ONLY,
                        "description": HapStatusCode.CANT_WRITE_READ_ONLY.description,
                    }

        return results

    async def subscribe(self, characteristics: Iterable[tuple[int, int]]) -> None:
        """Subscribe to characteristics."""
        new_chars = await super().subscribe(characteristics)
        if not new_chars or not self.client or not self.client.is_connected:
            # Don't force a new connection if we are not already
            # connected as we will get disconnected events.
            return
        # We do not want to block setup of the accessory so we
        # do not wait for the result as any failures will be
        # handled by the retry logic or fallback to disconnected
        # events.
        async_create_task(self._async_subscribe(new_chars))

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def _async_subscribe(self, new_chars: Iterable[tuple[int, int]]) -> None:
        """Subscribe to new characteristics."""
        logger.debug("%s: subscribing to %s", self.name, new_chars)
        await self._populate_accessories_and_characteristics()
        if not self._broadcast_decryption_key:
            await self._async_set_broadcast_encryption_key()
        await self._async_subscribe_broadcast_events(new_chars)
        await self._async_start_notify_subscriptions(new_chars)

    async def unsubscribe(self, characteristics: Iterable[tuple[int, int]]) -> None:
        pass

    async def identify(self):
        info = self.accessories.aid(BLE_AID).services.first(
            service_type=ServicesTypes.ACCESSORY_INFORMATION
        )
        char = info[CharacteristicsTypes.IDENTIFY]

        await self.put_characteristics(
            [
                (BLE_AID, char.iid, True),
            ]
        )

    @operation_lock
    @retry_bluetooth_connection_error()
    @restore_connection_and_resume
    async def add_pairing(
        self, additional_controller_pairing_identifier, ios_device_ltpk, permissions
    ):
        if permissions == "User":
            permissions = TLV.kTLVType_Permission_RegularUser
        elif permissions == "Admin":
            permissions = TLV.kTLVType_Permission_AdminUser
        else:
            raise RuntimeError(f"{self.name} Unknown permission: {permissions}")

        request_tlv = TLV.encode_list(
            [
                (TLV.kTLVType_State, TLV.M1),
                (TLV.kTLVType_Method, TLV.AddPairing),
                (
                    TLV.kTLVType_Identifier,
                    additional_controller_pairing_identifier.encode(),
                ),
                (TLV.kTLVType_PublicKey, bytes.fromhex(ios_device_ltpk)),
                (TLV.kTLVType_Permissions, permissions),
            ]
        )

        request_tlv = TLV.encode_list(
            [
                (TLV.kTLVHAPParamParamReturnResponse, bytearray(b"\x01")),
                (TLV.kTLVHAPParamValue, request_tlv),
            ]
        )

        info = self.accessories.aid(BLE_AID).services.first(
            service_type=ServicesTypes.PAIRING
        )
        char = info[CharacteristicsTypes.PAIRING_PAIRINGS]

        resp = await self._async_request(OpCode.CHAR_WRITE, char, request_tlv)

        response = dict(TLV.decode_bytes(resp))

        data = dict(TLV.decode_bytes(response[1]))

        if data.get(TLV.kTLVType_State, TLV.M2) != TLV.M2:
            raise InvalidError(
                f"{self.name}: Unexpected state after removing pairing request"
            )

        if TLV.kTLVType_Error in data:
            if data[TLV.kTLVType_Error] == TLV.kTLVError_Authentication:
                raise AuthenticationError(
                    f"{self.name}: Add pairing failed: insufficient access"
                )
            raise UnknownError(f"{self.name}: Add pairing failed: unknown error")

    @operation_lock
    @retry_bluetooth_connection_error(attempts=10)
    @restore_connection_and_resume
    async def remove_pairing(self, pairingId: str) -> bool:
        request_tlv = TLV.encode_list(
            [
                (TLV.kTLVType_State, TLV.M1),
                (TLV.kTLVType_Method, TLV.RemovePairing),
                (TLV.kTLVType_Identifier, pairingId.encode("utf-8")),
            ]
        )

        request_tlv = TLV.encode_list(
            [
                (TLV.kTLVHAPParamParamReturnResponse, bytearray(b"\x01")),
                (TLV.kTLVHAPParamValue, request_tlv),
            ]
        )

        info = self.accessories.aid(BLE_AID).services.first(
            service_type=ServicesTypes.PAIRING
        )
        char = info[CharacteristicsTypes.PAIRING_PAIRINGS]

        resp = await self._async_request(OpCode.CHAR_WRITE, char, request_tlv)

        response = dict(TLV.decode_bytes(resp))

        data = dict(TLV.decode_bytes(response[1]))

        if data.get(TLV.kTLVType_State, TLV.M2) != TLV.M2:
            raise InvalidError(
                f"{self.name}: Unexpected state after removing pairing request"
            )

        if TLV.kTLVType_Error in data:
            if data[TLV.kTLVType_Error] == TLV.kTLVError_Authentication:
                raise AuthenticationError(
                    f"{self.name}: Remove pairing failed: insufficient access"
                )
            raise UnknownError(f"{self.name}: Remove pairing failed: unknown error")

        await self._shutdown_if_primary_pairing_removed(pairingId)
        return True

    async def image(self, accessory: int, width: int, height: int) -> None:
        """Bluetooth devices don't return images."""
        return None



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implements the ChaCha20 stream cipher and the Poly1350 authenticator. More information can be found on
https://tools.ietf.org/html/rfc7539. See HomeKit spec page 51.
"""

from __future__ import annotations

import logging
import struct

from chacha20poly1305 import (
    ChaCha,
    ChaCha20Poly1305 as ChaCha20Poly1305PurePython,
    Poly1305,
)
from chacha20poly1305_reuseable import ChaCha20Poly1305Reusable
from cryptography.exceptions import InvalidTag

logger = logging.getLogger(__name__)


class ChaCha20Poly1305Encryptor:
    """Encrypt data with ChaCha20Poly1305."""

    def __init__(self, key: bytes) -> None:
        """Init the encryptor

        :param key: 256-bit (32-byte) key of type bytes
        """
        assert type(key) is bytes, "key is no instance of bytes"
        assert len(key) == 32
        self.chacha = ChaCha20Poly1305Reusable(key)

    def encrypt(
        self, aad: bytes, iv: bytes, constant: bytes, plaintext: bytes
    ) -> tuple[bytearray, bytes]:
        """
        The encrypt method for chacha20 aead as required by the Apple specification. The 96-bit nonce from RFC7539 is
        formed from the constant and the initialisation vector.

        :param aad: arbitrary length additional authenticated data
        :param iv: the initialisation vector
        :param constant: constant
        :param plaintext: arbitrary length plaintext of type bytes or bytearray
        :return: the cipher text and tag
        """
        assert type(plaintext) in [
            bytes,
            bytearray,
        ], "plaintext is no instance of bytes: %s" % str(type(plaintext))
        nonce = constant + iv
        return self.chacha.encrypt(bytes(nonce), bytes(plaintext), bytes(aad))


class ChaCha20Poly1305Decryptor:
    """Decrypt data with ChaCha20Poly1305."""

    def __init__(self, key: bytes) -> None:
        """Init the decrypter

        :param key: 256-bit (32-byte) key of type bytes
        """
        assert type(key) is bytes, "key is no instance of bytes"
        assert len(key) == 32
        self.chacha = ChaCha20Poly1305Reusable(key)

    def decrypt(
        self, aad: bytes, iv: bytes, constant: bytes, ciphertext: bytes
    ) -> bool | bytearray:
        """
        The decrypt method for chacha20 aead as required by the Apple specification. The 96-bit nonce from RFC7539 is
        formed from the constant and the initialisation vector.

        :param aad: arbitrary length additional authenticated data
        :param iv: the initialisation vector
        :param constant: constant
        :param ciphertext: arbitrary length plaintext of type bytes or bytearray
        :return: False if the tag could not be verified or the plaintext as bytes
        """
        assert type(ciphertext) in [
            bytes,
            bytearray,
        ], "ciphertext is no instance of bytes: %s" % str(type(ciphertext))
        nonce = constant + iv
        try:
            return bytearray(self.chacha.decrypt(nonce, bytes(ciphertext), bytes(aad)))
        except InvalidTag:
            # This should raise rather than the callees having to test for False
            return False


class ChaCha20Poly1305PartialTag(ChaCha20Poly1305PurePython):
    def open(self, nonce: bytes, combined_text: bytes, data: bytes) -> bytes:
        """
        Decrypts and authenticates ciphertext using nonce and data. If the
        tag is valid, the plaintext is returned. If the tag is invalid,
        returns None.

        This decryption only handles ble advertisements, which have a 4 byte
        partial tag.
        """
        if len(nonce) != 12:
            raise ValueError("Nonce must be 96 bit long")

        expected_tag = combined_text[-4:]
        ciphertext = combined_text[:-4]

        otk = self.poly1305_key_gen(self.key, nonce)

        mac_data = data + self.pad16(data)
        mac_data += ciphertext + self.pad16(ciphertext)
        mac_data += struct.pack("<Q", len(data))
        mac_data += struct.pack("<Q", len(ciphertext))
        tag = Poly1305(otk).create_tag(mac_data)

        if not tag.startswith(expected_tag):
            return None
        return ChaCha(self.key, nonce, counter=1).decrypt(ciphertext)



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import asyncio
from datetime import timedelta
from itertools import groupby
import logging
from operator import itemgetter
from typing import Any

from aiohomekit.controller.abstract import AbstractController, AbstractPairingData
from aiohomekit.exceptions import (
    AccessoryDisconnectedError,
    AuthenticationError,
    HttpErrorResponse,
    HttpException,
    InvalidError,
    UnknownError,
    UnpairedError,
)
import aiohomekit.hkjson as hkjson
from aiohomekit.http import HttpContentTypes
from aiohomekit.model import Accessories, AccessoriesState, Transport
from aiohomekit.model.characteristics import CharacteristicsTypes
from aiohomekit.protocol import error_handler
from aiohomekit.protocol.statuscodes import to_status_code
from aiohomekit.protocol.tlv import TLV
from aiohomekit.uuid import normalize_uuid
from aiohomekit.zeroconf import HomeKitService, ZeroconfPairing

from .connection import SecureHomeKitConnection

logger = logging.getLogger(__name__)


EMPTY_EVENT = {}


def format_characteristic_list(data):
    tmp = {}
    for c in data["characteristics"]:
        key = (c["aid"], c["iid"])
        del c["aid"]
        del c["iid"]

        if "status" in c and c["status"] == 0:
            del c["status"]
        if "status" in c and c["status"] != 0:
            c["description"] = to_status_code(c["status"]).description
        tmp[key] = c
    return tmp


class IpPairing(ZeroconfPairing):
    """
    This represents a paired HomeKit IP accessory.
    """

    def __init__(
        self, controller: AbstractController, pairing_data: AbstractPairingData
    ) -> None:
        """
        Initialize a Pairing by using the data either loaded from file or obtained after calling
        Controller.perform_pairing().

        :param pairing_data:
        """
        self.pairing_data = pairing_data
        self.connection = SecureHomeKitConnection(self, self.pairing_data)
        self.supports_subscribe = True

        super().__init__(controller, pairing_data)

    @property
    def is_connected(self) -> bool:
        return self.connection.is_connected

    @property
    def is_available(self) -> bool:
        """Returns true if the device is currently available."""
        return self.connection.is_connected

    @property
    def transport(self) -> Transport:
        """The transport used for the connection."""
        return Transport.IP

    @property
    def poll_interval(self) -> timedelta:
        """Returns how often the device should be polled."""
        return timedelta(minutes=1)

    @property
    def name(self) -> str:
        """Return the name of the pairing with the address."""
        if self.description:
            return f"{self.description.name} [{self.connection.host}:{self.connection.port}] (id={self.id})"
        return f"[{self.connection.host}:{self.connection.port}] (id={self.id})"

    def event_received(self, event):
        self._callback_listeners(format_characteristic_list(event))

    def _callback_listeners(self, event):
        for listener in self.listeners:
            try:
                listener(event)
            except Exception:
                logger.exception("Unhandled error when processing event")

    async def connection_made(self, secure):
        if not secure:
            return

        # Let our listeners know the connection is available again
        self._callback_listeners(EMPTY_EVENT)

        if self.subscriptions:
            await self.subscribe(self.subscriptions)

    async def _ensure_connected(self):
        if self._shutdown:
            return
        try:
            await asyncio.wait_for(self.connection.ensure_connection(), 10)
        except asyncio.TimeoutError:
            raise AccessoryDisconnectedError(
                f"Timeout while waiting for connection to device {self.connection.host}:{self.connection.port}"
            )

        if not self.connection.is_connected:
            raise AccessoryDisconnectedError(
                f"Ensure connection returned but still not connected: {self.connection.host}:{self.connection.port}"
            )

        else:
            self._callback_availability_changed(True)

    async def close(self) -> None:
        """
        Close the pairing's communications. This closes the session.
        """
        await self.connection.close()
        await asyncio.sleep(0)

    async def list_accessories_and_characteristics(self) -> list[dict[str, Any]]:
        """
        This retrieves a current set of accessories and characteristics behind this pairing.

        :return: the accessory data as described in the spec on page 73 and following
        :raises AccessoryNotFoundError: if the device can not be found via zeroconf
        """
        await self._ensure_connected()

        response = await self.connection.get_json("/accessories")

        accessories = response["accessories"]

        for accessory in accessories:
            for service in accessory["services"]:
                service["type"] = normalize_uuid(service["type"])

                for characteristic in service["characteristics"]:
                    characteristic["type"] = normalize_uuid(characteristic["type"])

        self._accessories_state = AccessoriesState(
            Accessories.from_list(accessories), self.config_num or 0
        )
        self._update_accessories_state_cache()
        return accessories

    async def list_pairings(self):
        """
        This method returns all pairings of a HomeKit accessory. This always includes the local controller and can only
        be done by an admin controller.

        The keys in the resulting dicts are:
         * pairingId: the pairing id of the controller
         * publicKey: the ED25519 long-term public key of the controller
         * permissions: bit value for the permissions
         * controllerType: either admin or regular

        :return: a list of dicts
        :raises: UnknownError: if it receives unexpected data
        :raises: UnpairedError: if the polled accessory is not paired
        """
        await self._ensure_connected()

        data = await self.connection.post_tlv(
            "/pairings",
            [(TLV.kTLVType_State, TLV.M1), (TLV.kTLVType_Method, TLV.ListPairings)],
        )

        if not (data[0][0] == TLV.kTLVType_State and data[0][1] == TLV.M2):
            raise UnknownError("unexpected data received: " + str(data))

        if (
            data[1][0] == TLV.kTLVType_Error
            and data[1][1] == TLV.kTLVError_Authentication
        ):
            raise UnpairedError("Must be paired")

        tmp = []
        r = {}
        for d in data[1:]:
            if d[0] == TLV.kTLVType_Identifier:
                r = {}
                tmp.append(r)
                r["pairingId"] = d[1].decode()
            if d[0] == TLV.kTLVType_PublicKey:
                r["publicKey"] = d[1].hex()
            if d[0] == TLV.kTLVType_Permissions:
                controller_type = "regular"
                if d[1] == b"\x01":
                    controller_type = "admin"
                r["permissions"] = int.from_bytes(d[1], byteorder="little")
                r["controllerType"] = controller_type
        return tmp

    async def get_characteristics(
        self,
        characteristics,
    ):
        """
        This method is used to get the current readouts of any characteristic of the accessory.

        :param characteristics: a list of 2-tupels of accessory id and instance id
        :param include_meta: if True, include meta information about the characteristics. This contains the format and
                             the various constraints like maxLen and so on.
        :param include_perms: if True, include the permissions for the requested characteristics.
        :param include_type: if True, include the type of the characteristics in the result. See CharacteristicsTypes
                             for translations.
        :param include_events: if True on a characteristics that supports events, the result will contain information if
                               the controller currently is receiving events for that characteristic. Key is 'ev'.
        :return: a dict mapping 2-tupels of aid and iid to dicts with value or status and description, e.g.
                 {(1, 8): {'value': 23.42}
                  (1, 37): {'description': 'Resource does not exist.', 'status': -70409}
                 }
        """
        await self._ensure_connected()

        if not self.accessories:
            await self.list_accessories_and_characteristics()

        url = "/characteristics?id=" + ",".join(
            str(x[0]) + "." + str(x[1]) for x in set(characteristics)
        )

        response = await self.connection.get_json(url)

        return format_characteristic_list(response)

    async def put_characteristics(self, characteristics):
        """
        Update the values of writable characteristics. The characteristics have to be identified by accessory id (aid),
        instance id (iid). If do_conversion is False (the default), the value must be of proper format for the
        characteristic since no conversion is done. If do_conversion is True, the value is converted.

        :param characteristics: a list of 3-tupels of accessory id, instance id and the value
        :param do_conversion: select if conversion is done (False is default)
        :return: a dict from (aid, iid) onto {status, description}
        :raises FormatError: if the input value could not be converted to the target type and conversion was
                             requested
        """
        await self._ensure_connected()

        if not self.accessories:
            await self.list_accessories_and_characteristics()

        data = []
        characteristics_set = set()
        for characteristic in characteristics:
            aid = characteristic[0]
            iid = characteristic[1]
            value = characteristic[2]
            characteristics_set.add(f"{aid}.{iid}")
            data.append({"aid": aid, "iid": iid, "value": value})
        data = {"characteristics": data}

        response = await self.connection.put_json("/characteristics", data)
        if response:
            data = {
                (d["aid"], d["iid"]): {
                    "status": d["status"],
                    "description": to_status_code(d["status"]).description,
                }
                for d in response["characteristics"]
            }
            return data

        return {}

    async def subscribe(self, characteristics):
        await super().subscribe(set(characteristics))

        if not self.supports_subscribe:
            logger.info(
                "This device does not support push, so only polling operations will be supported during this session"
            )
            return

        try:
            await self._ensure_connected()
        except AccessoryDisconnectedError:
            logger.debug(
                "Attempted to subscribe to characteristics but could not connect to accessory"
            )
            return {}

        try:
            return await self._update_subscriptions(characteristics, True)
        except AccessoryDisconnectedError:
            self.supports_subscribe = False
            return {}

    async def unsubscribe(self, characteristics):
        if not self.connection.is_connected:
            # If not connected no need to unsubscribe
            await super().unsubscribe(characteristics)
            return {}

        await self._ensure_connected()
        char_set = set(characteristics)
        status = await self._update_subscriptions(characteristics, False)
        for id_tuple in status:
            char_set.discard(id_tuple)

        await super().unsubscribe(char_set)
        return status

    async def _update_subscriptions(self, characteristics, ev):
        """Subscribe or unsubscribe to characteristics."""
        status = {}
        # We do one aid at a time to match what iOS does
        # even though its inefficient
        # https://github.com/home-assistant/core/issues/37996
        #
        # Prebuild the payloads to avoid the set size changing
        # between await calls
        char_payloads = [
            [{"aid": aid, "iid": iid, "ev": ev} for aid, iid in aid_iids]
            for _, aid_iids in groupby(characteristics, key=itemgetter(0))
        ]
        for char_payload in char_payloads:
            response = await self.connection.put_json(
                "/characteristics",
                {"characteristics": char_payload},
            )
            if response:
                # An empty body is a success response
                for row in response.get("characteristics", []):
                    status[(row["aid"], row["iid"])] = {
                        "status": row["status"],
                        "description": to_status_code(row["status"]).description,
                    }

        return status

    async def async_populate_accessories_state(
        self, force_update: bool = False, attempts: int | None = None
    ) -> bool:
        """Populate the state of all accessories.

        This method should try not to fetch all the accessories unless
        we know the config num is out of date or force_update is True
        """
        if not self.accessories or force_update:
            await self.list_accessories_and_characteristics()

    async def _process_config_changed(self, config_num: int) -> None:
        """Process a config change.

        This method is called when the config num changes.
        """
        await self.list_accessories_and_characteristics()
        self._accessories_state = AccessoriesState(
            self._accessories_state.accessories, config_num
        )
        self._callback_and_save_config_changed(self.config_num)

    async def _process_disconnected_events(self):
        """Process any events that happened while we were disconnected.

        We don't disconnect in IP so there is no need to do anything here.
        """

    async def identify(self):
        """
        This call can be used to trigger the identification of a paired accessory. A successful call should
        cause the accessory to perform some specific action by which it can be distinguished from the others (blink a
        LED for example).

        It uses the identify characteristic as described on page 152 of the spec.

        :return True, if the identification was run, False otherwise
        """
        await self._ensure_connected()

        if not self.accessories:
            await self.list_accessories_and_characteristics()

        # we are looking for a characteristic of the identify type
        identify_type = CharacteristicsTypes.IDENTIFY

        # search all accessories, all services and all characteristics
        logger.debug("Searching for identify characteristic in %s", self.accessories)
        for accessory in self.accessories:
            aid = accessory.aid
            for service in accessory.services:
                for characteristic in service.characteristics:
                    iid = characteristic.iid
                    c_type = normalize_uuid(characteristic.type)
                    if identify_type == c_type:
                        # found the identify characteristic, so let's put a value there
                        if not await self.put_characteristics([(aid, iid, True)]):
                            return True
        return False

    async def add_pairing(
        self, additional_controller_pairing_identifier, ios_device_ltpk, permissions
    ):
        await self._ensure_connected()

        if permissions == "User":
            permissions = TLV.kTLVType_Permission_RegularUser
        elif permissions == "Admin":
            permissions = TLV.kTLVType_Permission_AdminUser
        else:
            raise RuntimeError(f"Unknown permission: {permissions}")

        request_tlv = [
            (TLV.kTLVType_State, TLV.M1),
            (TLV.kTLVType_Method, TLV.AddPairing),
            (
                TLV.kTLVType_Identifier,
                additional_controller_pairing_identifier.encode(),
            ),
            (TLV.kTLVType_PublicKey, bytes.fromhex(ios_device_ltpk)),
            (TLV.kTLVType_Permissions, permissions),
        ]

        data = dict(await self.connection.post_tlv("/pairings", request_tlv))

        if data.get(TLV.kTLVType_State, TLV.M2) != TLV.M2:
            raise InvalidError("Unexpected state after add pairing request")

        if TLV.kTLVType_Error in data:
            error_handler(data[TLV.kTLVType_Error], "M2")

        return True

    async def remove_pairing(self, pairingId: str) -> bool:
        """
        Remove a pairing between the controller and the accessory. The pairing data is delete on both ends, on the
        accessory and the controller.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the accessory seems still
            to be paired on the next start of the application.

        :param alias: the controller's alias for the accessory
        :param pairingId: the pairing id to be removed
        :raises AuthenticationError: if the controller isn't authenticated to the accessory.
        :raises AccessoryNotFoundError: if the device can not be found via zeroconf
        :raises UnknownError: on unknown errors
        """
        await self._ensure_connected()

        request_tlv = [
            (TLV.kTLVType_State, TLV.M1),
            (TLV.kTLVType_Method, TLV.RemovePairing),
            (TLV.kTLVType_Identifier, pairingId.encode("utf-8")),
        ]

        data = dict(await self.connection.post_tlv("/pairings", request_tlv))

        if data.get(TLV.kTLVType_State, TLV.M2) != TLV.M2:
            raise InvalidError("Unexpected state after removing pairing request")

        if TLV.kTLVType_Error in data:
            if data[TLV.kTLVType_Error] == TLV.kTLVError_Authentication:
                raise AuthenticationError("Remove pairing failed: insufficient access")
            raise UnknownError("Remove pairing failed: unknown error")

        await self._shutdown_if_primary_pairing_removed(pairingId)
        return True

    async def image(self, accessory: int, width: int, height: int) -> bytes:
        await self._ensure_connected()

        try:
            resp = await self.connection.post(
                "/resource",
                content_type=HttpContentTypes.JSON,
                body=hkjson.dump_bytes(
                    {
                        "aid": accessory,
                        "resource-type": "image",
                        "image-width": width,
                        "image-height": height,
                    }
                ),
            )

        except HttpException:
            return None

        except HttpErrorResponse:
            return None

        except AccessoryDisconnectedError:
            return None

        return resp.body

    def _async_description_update(self, description: HomeKitService | None) -> None:
        """We have new zeroconf metadata for this device."""
        super()._async_description_update(description)

        # If we are not connected, or are in the process of reconnecting, hasten the process
        self.connection.reconnect_soon()



#
# Copyright 2022 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mechanism to cache characteristic database.

It is slow to query the BLE characteristics to find their iid and
signatures. We only need to do this work when the cn has incremented.

This interface must be kept compatible with Home Assistant. This is a
dumb implementation for development and CLI usage.
"""

from __future__ import annotations

import logging
import pathlib
from typing import Any, Protocol, TypedDict

import aiohomekit.hkjson as hkjson

logger = logging.getLogger(__name__)


class Pairing(TypedDict):
    """A versioned map of entity metadata as presented by aiohomekit."""

    config_num: int
    accessories: list[Any]
    broadcast_key: str | None


class StorageLayout(TypedDict):
    """Cached pairing metadata needed by aiohomekit."""

    pairings: dict[str, Pairing]


class CharacteristicCacheType(Protocol):
    def get_map(self, homekit_id: str) -> Pairing | None:
        pass

    def async_create_or_update_map(
        self,
        homekit_id: str,
        config_num: int,
        accessories: list[Any],
        broadcast_key: str | None = None,
    ) -> Pairing:
        pass

    def async_delete_map(self, homekit_id: str) -> None:
        pass


class CharacteristicCacheMemory:
    def __init__(self) -> None:
        """Create a new entity map store."""
        self.storage_data: dict[str, Pairing] = {}

    def get_map(self, homekit_id: str) -> Pairing | None:
        """Get a pairing cache item."""
        return self.storage_data.get(homekit_id)

    def async_create_or_update_map(
        self,
        homekit_id: str,
        config_num: int,
        accessories: list[Any],
        broadcast_key: str | None = None,
    ) -> Pairing:
        """Create a new pairing cache."""
        data = Pairing(
            config_num=config_num, accessories=accessories, broadcast_key=broadcast_key
        )
        self.storage_data[homekit_id] = data
        return data

    def async_delete_map(self, homekit_id: str) -> None:
        """Delete pairing cache."""
        if homekit_id not in self.storage_data:
            return

        self.storage_data.pop(homekit_id)


class CharacteristicCacheFile(CharacteristicCacheMemory):
    def __init__(self, location: pathlib.Path) -> None:
        """Create a new entity map store."""
        super().__init__()

        self.location = location
        if location.exists():
            with open(location, encoding="utf-8") as fp:
                try:
                    self.storage_data = hkjson.loads(fp.read())["pairings"]
                except hkjson.JSON_DECODE_EXCEPTIONS:
                    logger.debug(
                        "Characteristic cache was corrupted, proceeding with cold cache"
                    )

    def async_create_or_update_map(
        self,
        homekit_id: str,
        config_num: int,
        accessories: list[Any],
        broadcast_key: bytes | None = None,
    ) -> Pairing:
        """Create a new pairing cache."""
        data = super().async_create_or_update_map(
            homekit_id, config_num, accessories, broadcast_key
        )
        self._do_save()
        return data

    def async_delete_map(self, homekit_id: str) -> None:
        """Delete pairing cache."""
        super().async_delete_map(homekit_id)
        self._do_save()

    def _do_save(self) -> None:
        """Schedule saving the entity map cache."""
        with open(self.location, mode="w", encoding="utf-8") as fp:
            fp.write(hkjson.dumps(self._data_to_save()))

    def _data_to_save(self) -> dict[str, Any]:
        """Return data of entity map to store in a file."""
        return StorageLayout(pairings=self.storage_data)
